import math
import torch
import unittest
from tests.utils import generate_simple_hete_graph
from tests.utils import generate_simple_hete_dataset
from tests.utils import generate_dense_hete_graph
from tests.utils import generate_dense_hete_dataset
from tests.utils import generate_dense_hete_multigraph
from deepsnap.graph import Graph
from deepsnap.batch import Batch
from deepsnap.dataset import GraphDataset
from torch.utils.data import DataLoader
from deepsnap.hetero_graph import HeteroGraph


class TestHeteroGraph(unittest.TestCase):
    def test_hetero_graph_basics(self):
        G = generate_simple_hete_graph()
        hete = HeteroGraph(G)

        self.assertEqual(hete.get_num_node_features('n1'), 10)
        self.assertEqual(hete.get_num_node_features('n2'), 12)
        self.assertEqual(hete.get_num_edge_features('e1'), 8)
        self.assertEqual(hete.get_num_edge_features('e2'), 12)
        self.assertEqual(hete.get_num_nodes('n1'), 4)
        self.assertEqual(hete.get_num_nodes('n2'), 5)
        self.assertEqual(len(hete.node_types), 2)
        self.assertEqual(len(hete.edge_types), 2)

        message_types = hete.message_types
        self.assertEqual(len(message_types), 7)
        self.assertEqual(hete.get_num_node_labels('n1'), 2)
        self.assertEqual(hete.get_num_node_labels('n2'), 2)
        self.assertEqual(hete.get_num_edge_labels('e1'), 3)
        self.assertEqual(hete.get_num_edge_labels('e2'), 3)
        self.assertEqual(hete.get_num_edges(message_types[0]), 3)
        self.assertEqual(len(hete.node_label_index), 2)

    def test_hetero_graph_batch(self):
        G = generate_simple_hete_graph()
        hete = HeteroGraph(G)

        heteGraphDataset = []
        for i in range(30):
            heteGraphDataset.append(hete.clone())
        dataloader = DataLoader(
                        heteGraphDataset, collate_fn=Batch.collate(),
                        batch_size=3, shuffle=True)

        self.assertEqual(len(dataloader), math.ceil(30 / 3))
        for data in dataloader:
            self.assertEqual(data.num_graphs, 3)

    def test_hetero_multigraph_split(self):
        G = generate_dense_hete_multigraph()
        hete = HeteroGraph(G)

        # node
        hete_node = hete.split(task='node')
        for node_type in hete.node_label_index:
            num_nodes = len(hete.node_label_index[node_type])
            num_nodes_reduced = num_nodes - 3
            node_0 = 1 + int(num_nodes_reduced * 0.8)
            node_1 = 1 + int(num_nodes_reduced * 0.1)
            node_2 = num_nodes - node_0 - node_1
            self.assertEqual(
                len(hete_node[0].node_label_index[node_type]), node_0)
            self.assertEqual(
                len(hete_node[1].node_label_index[node_type]), node_1)
            self.assertEqual(
                len(hete_node[2].node_label_index[node_type]), node_2)

        # edge
        hete_edge = hete.split(task='edge')
        for edge_type in hete.edge_label_index:
            num_edges = int(hete.edge_label_index[edge_type].shape[1])
            num_edges_reduced = num_edges - 3
            edge_0 = 1 + int(num_edges_reduced * 0.8)
            edge_1 = 1 + int(num_edges_reduced * 0.1)
            edge_2 = num_edges - edge_0 - edge_1
            self.assertEqual(
                hete_edge[0].edge_label_index[edge_type].shape[1], edge_0)
            self.assertEqual(
                hete_edge[1].edge_label_index[edge_type].shape[1], edge_1)
            self.assertEqual(
                hete_edge[2].edge_label_index[edge_type].shape[1], edge_2)

        # link prediction
        hete_link = hete.split(task='link_pred', split_ratio=[0.5, 0.3, 0.2])
        # calculate the expected edge num for each splitted subgraph
        hete_link_train_edge_num, hete_link_val_edge_num, hete_link_test_edge_num = 0, 0, 0
        for key, val in hete.edge_label_index.items():
            val_length = val.shape[1]
            val_length_reduced = val_length - 3
            hete_link_train_edge_num += 1 + int(0.5 * val_length_reduced)
            hete_link_val_edge_num += 1 + int(0.3 * val_length_reduced)
            hete_link_test_edge_num += \
                val_length - 2 - int(0.5 * val_length_reduced) - int(0.3 * val_length_reduced)

        self.assertEqual(len(hete_link[0].edge_label), hete_link_train_edge_num)
        self.assertEqual(len(hete_link[1].edge_label), hete_link_val_edge_num)
        self.assertEqual(len(hete_link[2].edge_label), hete_link_test_edge_num)

    def test_hetero_graph_split(self):
        G = generate_dense_hete_graph()
        hete = HeteroGraph(G)

        hete_node = hete.split()
        for node_type in hete.node_label_index:
            num_nodes = len(hete.node_label_index[node_type])
            num_nodes_reduced = num_nodes - 3
            node_0 = 1 + int(num_nodes_reduced * 0.8)
            node_1 = 1 + int(num_nodes_reduced * 0.1)
            node_2 = num_nodes - node_0 - node_1
            self.assertEqual(
                len(hete_node[0].node_label_index[node_type]), node_0)
            self.assertEqual(
                len(hete_node[1].node_label_index[node_type]), node_1)
            self.assertEqual(
                len(hete_node[2].node_label_index[node_type]), node_2)

        # node with specified split type
        node_split_types = ['n1']
        hete_node = hete.split(split_types=node_split_types)
        for node_type in hete.node_label_index:
            if (node_type in node_split_types):
                num_nodes = len(hete.node_label_index[node_type])
                num_nodes_reduced = num_nodes - 3
                node_0 = 1 + int(num_nodes_reduced * 0.8)
                node_1 = 1 + int(num_nodes_reduced * 0.1)
                node_2 = num_nodes - node_0 - node_1
                self.assertEqual(
                    len(hete_node[0].node_label_index[node_type]), node_0)
                self.assertEqual(
                    len(hete_node[1].node_label_index[node_type]), node_1)
                self.assertEqual(
                    len(hete_node[2].node_label_index[node_type]), node_2)
            else:
                self.assertEqual(
                    len(hete_node[0].node_label_index[node_type]),
                    len(hete.node_label_index[node_type]))
                self.assertEqual(
                    len(hete_node[1].node_label_index[node_type]),
                    len(hete.node_label_index[node_type]))
                self.assertEqual(
                    len(hete_node[2].node_label_index[node_type]),
                    len(hete.node_label_index[node_type]))

        # edge
        hete_edge = hete.split(task='edge')
        for edge_type in hete.edge_label_index:
            num_edges = int(hete.edge_label_index[edge_type].shape[1])
            num_edges_reduced = num_edges - 3
            edge_0 = 1 + int(num_edges_reduced * 0.8)
            edge_1 = 1 + int(num_edges_reduced * 0.1)
            edge_2 = num_edges - edge_0 - edge_1
            self.assertEqual(
                hete_edge[0].edge_label_index[edge_type].shape[1], edge_0)
            self.assertEqual(
                hete_edge[1].edge_label_index[edge_type].shape[1], edge_1)
            self.assertEqual(
                hete_edge[2].edge_label_index[edge_type].shape[1], edge_2)

        # edge with specified split type
        edge_split_types = [('n1', 'e1', 'n1'), ('n1', 'e2', 'n2')]
        hete_edge = hete.split(task='edge', split_types=edge_split_types)
        for edge_type in hete.edge_label_index:
            if (edge_type in edge_split_types):
                num_edges = int(hete.edge_label_index[edge_type].shape[1])
                num_edges_reduced = num_edges - 3
                edge_0 = 1 + int(num_edges_reduced * 0.8)
                edge_1 = 1 + int(num_edges_reduced * 0.1)
                edge_2 = num_edges - edge_0 - edge_1

                self.assertEqual(
                    hete_edge[0].edge_label_index[edge_type].shape[1], edge_0)
                self.assertEqual(
                    hete_edge[1].edge_label_index[edge_type].shape[1], edge_1)
                self.assertEqual(
                    hete_edge[2].edge_label_index[edge_type].shape[1], edge_2)
            else:
                self.assertEqual(
                    hete_edge[0].edge_label_index[edge_type].shape[1],
                    hete.edge_label_index[edge_type].shape[1])
                self.assertEqual(
                    hete_edge[1].edge_label_index[edge_type].shape[1],
                    hete.edge_label_index[edge_type].shape[1])
                self.assertEqual(
                    hete_edge[2].edge_label_index[edge_type].shape[1],
                    hete.edge_label_index[edge_type].shape[1])

        # link_pred
        hete_link = hete.split(task='link_pred', split_ratio=[0.5, 0.3, 0.2])
        # calculate the expected edge num for each splitted subgraph
        hete_link_train_edge_num, hete_link_val_edge_num, hete_link_test_edge_num = 0, 0, 0
        for key, val in hete.edge_label_index.items():
            val_length = val.shape[1]
            val_length_reduced = val_length - 3
            hete_link_train_edge_num += 1 + int(0.5 * val_length_reduced)
            hete_link_val_edge_num += 1 + int(0.3 * val_length_reduced)
            hete_link_test_edge_num += \
                val_length - 2 - int(0.5 * val_length_reduced) - int(0.3 * val_length_reduced)

        self.assertEqual(len(hete_link[0].edge_label),
                         hete_link_train_edge_num)
        self.assertEqual(len(hete_link[1].edge_label),
                         hete_link_val_edge_num)
        self.assertEqual(len(hete_link[2].edge_label),
                         hete_link_test_edge_num)

    def test_hetero_graph_none(self):
        G = generate_simple_hete_graph(no_edge_type=True)
        hete = HeteroGraph(G)
        message_types = hete.message_types
        for message_type in message_types:
            self.assertEqual(message_type[1], None)


if __name__ == '__main__':
    unittest.main()



import numpy as np
import networkx as nx
import torch
import itertools
np.random.seed(0)


def simple_networkx_graph(directed=True):
    num_nodes = 10
    # node = np.arange(num_nodes).astype(np.int64)
    edge_index = \
        torch.tensor([[0, 0, 1, 1, 2, 2, 3, 3, 3, 4, 4, 5, 6, 6, 7, 7, 9],
                      [1, 2, 2, 3, 3, 8, 4, 5, 6, 5, 6, 7, 8, 9, 8, 9, 8]]).long()
    x = torch.zeros([num_nodes, 2])
    y = torch.tensor([0, 0, 1, 1, 2, 2, 3, 3, 4, 4]).long()
    for i in range(num_nodes):
        x[i] = np.random.randint(1, num_nodes)
    edge_x = torch.zeros([edge_index.shape[1], 2])
    edge_y = torch.tensor([0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3]).long()
    for i in range(edge_index.shape[1]):
        edge_x[i] = np.random.randint(1, num_nodes)

    G = nx.DiGraph()
    G.add_nodes_from(range(num_nodes))
    for i, (u, v) in enumerate(edge_index.T.tolist()):
        G.add_edge(u, v)

    # if it is undirected, modify the edge attributes
    if directed is False:
        G = G.to_undirected()
        H = G.to_directed()
        edge_index = np.zeros([2, edge_index.shape[1] * 2]).astype(np.int64)
        edge_x = np.zeros([edge_x.shape[0] * 2, edge_x.shape[1]])
        edge_y = np.zeros(edge_y.shape[0] * 2).astype(np.int64)
        for i, nx_edge in enumerate(nx.to_edgelist(H)):
            edge_index[:, i] = np.array([nx_edge[0], nx_edge[1]]).astype(np.int64)
            edge_x[i] = nx_edge[2]['edge_attr']
            edge_y[i] = nx_edge[2]['edge_y']

    graph_x = torch.tensor([[0, 1]])
    graph_y = torch.tensor([0])
    return G, x, y, edge_x, edge_y, edge_index, graph_x, graph_y


def sample_neigh(graph, size):
    while True:
        start_node = np.random.choice(list(graph.nodes))
        neigh = [start_node]
        frontier = list(set(graph.neighbors(start_node)) - set(neigh))
        visited = set([start_node])
        while len(neigh) < size and frontier:
            new_node = np.random.choice(list(frontier))
            assert new_node not in neigh
            neigh.append(new_node)
            visited.add(new_node)
            frontier += list(graph.neighbors(new_node))
            frontier = [x for x in frontier if x not in visited]
        if len(neigh) == size:
            return graph, neigh


def gen_graph(size, graph):
    graph, neigh = sample_neigh(graph, size)
    return graph.subgraph(neigh)


def generate_simple_hete_graph(no_edge_type=False):
    G = nx.DiGraph()
    for i in range(9):
        if i < 2:
            node_feature = torch.rand([10, ])
            node_type = "n1"
            node_label = 0
            G.add_node(i, node_type=node_type, node_label=node_label, node_feature=node_feature)
        elif i >=2 and i < 4:
            node_feature = torch.rand([12, ])
            node_type = "n2"
            node_label = 0
            G.add_node(i, node_type=node_type, node_label=node_label, node_feature=node_feature)
        elif i >=4 and i < 6:
            node_feature = torch.rand([10, ])
            node_type = "n1"
            node_label = 1
            G.add_node(i, node_type=node_type, node_label=node_label, node_feature=node_feature)
        else:
            node_feature = torch.rand([12, ])
            node_type = "n2"
            node_label = 1
            G.add_node(i, node_type=node_type, node_label=node_label, node_feature=node_feature)
    if no_edge_type:
        G.add_edge(0, 1, edge_label=0, edge_feature=torch.rand([8, ]))
        G.add_edge(0, 2, edge_label=1, edge_feature=torch.rand([8, ]))
        G.add_edge(0, 5, edge_label=0, edge_feature=torch.rand([8, ]))
        G.add_edge(1, 3, edge_label=0, edge_feature=torch.rand([8, ]))
        G.add_edge(1, 5, edge_label=1, edge_feature=torch.rand([8, ]))
        G.add_edge(2, 3, edge_label=1, edge_feature=torch.rand([8, ]))
        G.add_edge(2, 4, edge_label=2, edge_feature=torch.rand([8, ]))
        G.add_edge(3, 4, edge_label=2, edge_feature=torch.rand([8, ]))
        G.add_edge(4, 0, edge_label=1, edge_feature=torch.rand([8, ]))
        G.add_edge(4, 5, edge_label=1, edge_feature=torch.rand([8, ]))
        G.add_edge(5, 7, edge_label=1, edge_feature=torch.rand([8, ]))
        G.add_edge(6, 1, edge_label=1, edge_feature=torch.rand([8, ]))
        G.add_edge(6, 2, edge_label=1, edge_feature=torch.rand([8, ]))
        G.add_edge(7, 3, edge_label=2, edge_feature=torch.rand([8, ]))
        G.add_edge(8, 0, edge_label=0, edge_feature=torch.rand([8, ]))
        G.add_edge(8, 1, edge_label=0, edge_feature=torch.rand([8, ]))
    else:
        G.add_edge(0, 1, edge_label=0, edge_feature=torch.rand([8, ]), edge_type="e1")
        G.add_edge(0, 2, edge_label=1, edge_feature=torch.rand([12, ]), edge_type="e2")
        G.add_edge(0, 5, edge_label=0, edge_feature=torch.rand([8, ]), edge_type="e1")
        G.add_edge(1, 3, edge_label=0, edge_feature=torch.rand([12, ]), edge_type="e2")
        G.add_edge(1, 5, edge_label=1, edge_feature=torch.rand([12, ]), edge_type="e2")
        G.add_edge(2, 3, edge_label=1, edge_feature=torch.rand([8, ]), edge_type="e1")
        G.add_edge(2, 4, edge_label=2, edge_feature=torch.rand([12, ]), edge_type="e2")
        G.add_edge(3, 4, edge_label=2, edge_feature=torch.rand([12, ]), edge_type="e2")
        G.add_edge(4, 0, edge_label=1, edge_feature=torch.rand([12, ]), edge_type="e2")
        G.add_edge(4, 5, edge_label=1, edge_feature=torch.rand([8, ]), edge_type="e1")
        G.add_edge(5, 7, edge_label=1, edge_feature=torch.rand([8, ]), edge_type="e1")
        G.add_edge(6, 1, edge_label=1, edge_feature=torch.rand([8, ]), edge_type="e1")
        G.add_edge(6, 2, edge_label=1, edge_feature=torch.rand([8, ]), edge_type="e1")
        G.add_edge(7, 3, edge_label=2, edge_feature=torch.rand([8, ]), edge_type="e1")
        G.add_edge(8, 0, edge_label=0, edge_feature=torch.rand([12, ]), edge_type="e2")
        G.add_edge(8, 1, edge_label=0, edge_feature=torch.rand([12, ]), edge_type="e2")
    return G


def generate_simple_hete_dataset(no_edge_type=False):
    G = nx.DiGraph()
    for i in range(9):
        if i < 2:
            node_feature = torch.rand([10, ])
            node_type = "n1"
            node_label = 0
            G.add_node(i, node_type=node_type, node_label=node_label, node_feature=node_feature)
        elif i >= 2 and i < 4:
            node_feature = torch.rand([12, ])
            node_type = "n2"
            node_label = 0
            G.add_node(i, node_type=node_type, node_label=node_label, node_feature=node_feature)
        elif i >= 4 and i < 6:
            node_feature = torch.rand([10, ])
            node_type = "n1"
            node_label = 1
            G.add_node(i, node_type=node_type, node_label=node_label, node_feature=node_feature)
        else:
            node_feature = torch.rand([12, ])
            node_type = "n2"
            node_label = 1
            G.add_node(i, node_type=node_type, node_label=node_label, node_feature=node_feature)
    if no_edge_type:
        G.add_edge(0, 1, edge_feature=torch.rand([8, ]))
        G.add_edge(0, 2, edge_feature=torch.rand([8, ]))
        G.add_edge(0, 5, edge_feature=torch.rand([8, ]))
        G.add_edge(1, 3, edge_feature=torch.rand([8, ]))
        G.add_edge(1, 5, edge_feature=torch.rand([8, ]))
        G.add_edge(2, 3, edge_feature=torch.rand([8, ]))
        G.add_edge(2, 4, edge_feature=torch.rand([8, ]))
        G.add_edge(3, 4, edge_feature=torch.rand([8, ]))
        G.add_edge(4, 0, edge_feature=torch.rand([8, ]))
        G.add_edge(4, 5, edge_feature=torch.rand([8, ]))
        G.add_edge(5, 7, edge_feature=torch.rand([8, ]))
        G.add_edge(6, 1, edge_feature=torch.rand([8, ]))
        G.add_edge(6, 2, edge_feature=torch.rand([8, ]))
        G.add_edge(7, 3, edge_feature=torch.rand([8, ]))
        G.add_edge(8, 0, edge_feature=torch.rand([8, ]))
        G.add_edge(8, 1, edge_feature=torch.rand([8, ]))
    else:
        G.add_edge(0, 1, edge_feature=torch.rand([8, ]), edge_type="e1")
        G.add_edge(0, 2, edge_feature=torch.rand([12, ]), edge_type="e2")
        G.add_edge(0, 5, edge_feature=torch.rand([8, ]), edge_type="e1")
        G.add_edge(1, 3, edge_feature=torch.rand([12, ]), edge_type="e2")
        G.add_edge(1, 5, edge_feature=torch.rand([12, ]), edge_type="e2")
        G.add_edge(2, 3, edge_feature=torch.rand([8, ]), edge_type="e1")
        G.add_edge(2, 4, edge_feature=torch.rand([12, ]), edge_type="e2")
        G.add_edge(3, 4, edge_feature=torch.rand([12, ]), edge_type="e2")
        G.add_edge(4, 0, edge_feature=torch.rand([12, ]), edge_type="e2")
        G.add_edge(4, 5, edge_feature=torch.rand([8, ]), edge_type="e1")
        G.add_edge(5, 7, edge_feature=torch.rand([8, ]), edge_type="e1")
        G.add_edge(6, 1, edge_feature=torch.rand([8, ]), edge_type="e1")
        G.add_edge(6, 2, edge_feature=torch.rand([8, ]), edge_type="e1")
        G.add_edge(7, 3, edge_feature=torch.rand([8, ]), edge_type="e1")
        G.add_edge(8, 0, edge_feature=torch.rand([12, ]), edge_type="e2")
        G.add_edge(8, 1, edge_feature=torch.rand([12, ]), edge_type="e2")
    return G


def generate_dense_hete_graph(no_edge_type=False):
    G = nx.DiGraph()
    num_node = 20
    for i in range(num_node):
        if i < 10:
            node_feature = torch.rand([10, ])
            node_type = "n1"
            node_label = 0
            G.add_node(i, node_type=node_type, node_label=node_label, node_feature=node_feature)
        else:
            node_feature = torch.rand([12, ])
            node_type = "n2"
            node_label = 1
            G.add_node(i, node_type=node_type, node_label=node_label, node_feature=node_feature)

    if no_edge_type:
        for i, j in itertools.permutations(range(num_node), 2):
            rand = np.random.random()
            if (rand > 0.8):
                continue
            elif rand > 0.4:
                G.add_edge(i, j, edge_label=0, edge_feature=torch.rand([8, ]))
            else:
                G.add_edge(i, j, edge_label=0, edge_feature=torch.rand([8, ]))
    else:
        for i, j in itertools.permutations(range(num_node), 2):
            rand = np.random.random()
            if (rand > 0.8):
                continue
            elif rand > 0.4:
                G.add_edge(i, j, edge_label=0, edge_feature=torch.rand([8, ]), edge_type='e1')
            else:
                G.add_edge(i, j, edge_label=0, edge_feature=torch.rand([8, ]), edge_type='e2')
    return G


def generate_dense_hete_dataset(no_edge_type=False):
    G = nx.DiGraph()
    num_node = 20
    for i in range(num_node):
        if i < 10:
            node_feature = torch.rand([1, ])
            node_type = "n1"
            node_label = 0
            G.add_node(i, node_type=node_type, node_label=node_label, node_feature=node_feature)
        else:
            node_feature = torch.rand([1, ])
            node_type = "n2"
            node_label = 1
            G.add_node(i, node_type=node_type, node_label=node_label, node_feature=node_feature)

    if no_edge_type:
        for i, j in itertools.permutations(range(num_node), 2):
            rand = np.random.random()
            if (rand > 0.8):
                continue
            elif rand > 0.4:
                G.add_edge(i, j, edge_feature=torch.rand([1, ]))
            else:
                G.add_edge(i, j, edge_feature=torch.rand([1, ]))
    else:
        for i, j in itertools.permutations(range(num_node), 2):
            rand = np.random.random()
            if (rand > 0.8):
                continue
            elif rand > 0.4:
                G.add_edge(i, j, edge_feature=torch.rand([1, ]), edge_type='e1')
            else:
                G.add_edge(i, j, edge_feature=torch.rand([1, ]), edge_type='e2')
    return G


def generate_dense_hete_multigraph(no_edge_type=False):
    G = nx.MultiDiGraph()
    num_node = 20
    for i in range(num_node):
        if i < 10:
            node_feature = torch.rand([10, ])
            node_type = "n1"
            node_label = 0
            G.add_node(i, node_type=node_type, node_label=node_label, node_feature=node_feature)
        else:
            node_feature = torch.rand([12, ])
            node_type = "n2"
            node_label = 1
            G.add_node(i, node_type=node_type, node_label=node_label, node_feature=node_feature)

    if no_edge_type:
        for i, j in itertools.permutations(range(num_node), 2):
            rand = np.random.random()
            if (rand > 0.8):
                continue
            elif rand > 0.4:
                G.add_edge(i, j, edge_label=0, edge_feature=torch.rand([8, ]))
                G.add_edge(i, j, edge_label=0, edge_feature=torch.rand([8, ]))
            else:
                G.add_edge(i, j, edge_label=0, edge_feature=torch.rand([8, ]))
                G.add_edge(i, j, edge_label=0, edge_feature=torch.rand([8, ]))
    else:
        for i, j in itertools.permutations(range(num_node), 2):
            rand = np.random.random()
            if (rand > 0.8):
                continue
            elif rand > 0.4:
                G.add_edge(i, j, edge_label=0, edge_feature=torch.rand([8, ]), edge_type='e1')
                G.add_edge(i, j, edge_label=0, edge_feature=torch.rand([8, ]), edge_type='e1')
            else:
                G.add_edge(i, j, edge_label=0, edge_feature=torch.rand([8, ]), edge_type='e2')
                G.add_edge(i, j, edge_label=0, edge_feature=torch.rand([8, ]), edge_type='e2')
    return G



import math
import copy
import random
import torch
import networkx as nx
import numpy as np
from deepsnap.graph import Graph
from typing import *
import warnings


class HeteroGraph(Graph):
    r"""
    A plain python object modeling a heterogeneous graph with various
    attributes (String node type is required for the HeteroGraph).

    Args:
        G (:class:`networkx.classes.graph`): The NetworkX graph object which contains features
             and labels for each node type of edge type.
        **kwargs: keyworded argument list with keys such as :obj:`"node_feature"`, :obj:`"node_label"`
            and corresponding attributes.
    """
    def __init__(self, G=None, **kwargs):
        super(HeteroGraph, self).__init__()
        self.G = G
        if G is not None:
            keys = ['node_feature', 'node_label', 'edge_feature', 'edge_label',
                    'graph_feature', 'graph_label',
                    'edge_index', 'edge_label_index', 'node_label_index']
            for key in keys:
                self[key] = None

            for key, item in kwargs.items():
                self[key] = item

            self._update_tensors(init=True)
        self._num_positive_examples = None

    @property
    def node_types(self):
        r"""
        Return list of node types in the heterogeneous graph.
        """
        return list(self['node_feature'].keys())

    @property
    def edge_types(self):
        r"""
        Return list of edge types in the heterogeneous graph.
        """
        edge_type_set = set()
        for _, edge_type, _ in self['edge_index'].keys():
            edge_type_set.add(edge_type)
        return list(edge_type_set)

    @property
    def message_types(self):
        r"""
        Return the list of message types `(src_node_type, edge_type, end_node_type)`
        in the heterogeneous graph.
        """
        return list(self['edge_index'].keys())

    def get_num_nodes(self, node_type: Union[str, List[str]] = None):
        r"""
        Return number of nodes for a node type or list of node types.

        Args:
            node_type (str or list): Specified node type(s).

        Returns:
            int or list: The number of nodes for a node type or list of node types.
        """
        if 'node_feature' not in self:
            raise ValueError('Node feature is not available.')
        if node_type is None:
            node_type = self.node_types
        if isinstance(node_type, str):
            if node_type in self['node_feature']:
                return self['node_feature'][node_type].size(0)
            else:
                raise ValueError('Node type does not exist in stored node feature.')
        if isinstance(node_type, list):
            if not all(isinstance(node_type_i, str) for node_type_i in node_type):
                raise ValueError('Node type must be string.')
            if not all(node_type_i in self['node_feature'] for node_type_i in node_type):
                raise ValueError('Some node types do not exist in stored node feature.')
            else:
                num_nodes_dict = {}
                for node_type_i in node_type:
                    num_nodes_dict[node_type_i] = self['node_feature'][node_type_i].size(0)
                return num_nodes_dict
        else:
            raise TypeError('Node types must be string or list of strings.')

    def get_num_node_features(self, node_type: str) -> int:
        r"""
        Return the node feature dimension of specified node type.

        Returns:
            int: The node feature dimension for specified node type.
        """
        if 'node_feature' not in self or node_type not in self['node_feature']:
            return 0
        return self.get_num_dims('node_feature', node_type, as_label=False)

    def get_num_node_labels(self, node_type: str) -> int:
        r"""
        Return the number of node labels.

        Returns:
            int: Number of node labels for specified node type.
        """
        if 'node_label' not in self or node_type not in self['node_label']:
            return 0
        return self.get_num_dims('node_label', node_type, as_label=True)

    def get_num_edges(self, message_type: Union[tuple, List[tuple]] = None) -> int:
        r"""
        Return number of edges for a edge type or list of edgs types.

        Args:
            edge_type (str or list): Specified edge type(s).

        Returns:
            int or list: The number of edges for a edge type or list of edge types.
        """

        if 'edge_index' not in self:
            raise ValueError('Edge indices is not available')
        if message_type is None:
            message_type = self.message_types
        if isinstance(message_type, tuple):
            if message_type in self['edge_index']:
                return self['edge_index'][message_type].size(1)
            else:
                raise ValueError('Edge type does not exist in stored edge feature.')
        if isinstance(message_type, list):
            if not all(isinstance(message_type_i, tuple)
                       for message_type_i in message_type):
                raise ValueError('Edge type must be tuple.')
            if not all(message_type_i in self['edge_index']
                       for message_type_i in message_type):
                raise ValueError('Some edge types do not exist in stored edge feature.')
            else:
                num_edges_dict = {}
                for message_type_i in message_type:
                    num_edges_dict[message_type_i] = \
                        self['edge_index'][message_type_i].size(1)
                return num_edges_dict
        else:
            raise TypeError('Edge type must be tuple or list of tuple')

    def get_num_edge_labels(self, edge_type: str) -> int:
        r"""
        Return the number of edge labels.

        Returns:
            int: Number of edge labels for specified edge type.
        """
        if 'edge_label' not in self or edge_type not in self['edge_label']:
            return 0
        return self.get_num_dims('edge_label', edge_type, as_label=True)

    def get_num_edge_features(self, edge_type: str) -> int:
        r"""
        Return the edge feature dimension of specified edge type.

        Returns:
            int: The edge feature dimension for specified edge type.
        """
        if 'edge_feature' not in self or edge_type not in self['edge_feature']:
            return 0
        return self.get_num_dims('edge_feature', edge_type, as_label=False)

    def _get_node_type(self, node_dict: Dict):
        r"""
        Returns the node type of a node in its dict.

        Args:
            node_dict (dictionary): The node dictionary.

        Returns:
            the string of the node type: node type.
        """
        if 'node_type' not in node_dict:
            return None
        return node_dict['node_type']

    def _get_edge_type(self, edge_dict: Dict):
        r"""
        Similar to the `_get_node_type`
        """
        if 'edge_type' not in edge_dict:
            return None
        return edge_dict['edge_type']

    def _convert_to_graph_index(
            self,
            index: int,
            obj_type,
            mapping_type='node'):
        r"""
        Reverse operation of `_convert_to_tensor_index`
        """
        if mapping_type == 'node':
            mapping = self.node_to_graph_mapping
        elif mapping_type == "edge":
            mapping = self.edge_to_graph_mapping
        else:
            raise ValueError("Mapping type should be node or edge.")
        if obj_type not in mapping:
            raise ValueError("Node/edge type not in the graph.")
        return torch.index_select(mapping[obj_type], 0, index)

    def _convert_to_tensor_index(self, index: int, mapping_type="node"):
        r"""
        Returns specified type of index tensor.

        Args:
            index(tensor): the index tensor you want to transform.
            mapping_type(string): specify which mapping node or edge.

        Returns:
            index tensor.
        """
        if mapping_type == "node":
            mapping = self.node_to_tensor_mapping
        elif mapping_type == "edge":
            mapping = self.edge_to_tensor_mapping
        else:
            raise ValueError("Mapping type should be node or edge.")
        return torch.index_select(mapping, 0, index)

    def _get_node_attributes(self, key: str):
        r"""
        Returns the node attributes in the graph. Multiple attributes will be stacked.

        Args:
            key(string): the name of the attributes to return.

        Returns:
            a dictionary of node type to torch.tensor: node attributes.
        """
        attributes = {}
        indices = None
        if key == "node_feature":
            indices = {}

        for i, (_, node_dict) in enumerate(self.G.nodes.items()):
            if key in node_dict:
                node_type = self._get_node_type(node_dict)
                if node_type not in attributes:
                    attributes[node_type] = []
                    if indices is not None:
                        indices[node_type] = []
                attributes[node_type].append(node_dict[key])
                if indices is not None:
                    # use range(0 ~ num_nodes) as the graph node indices
                    indices[node_type].append(i)

        if len(attributes) == 0:
            return None

        for k, v in attributes.items():
            if torch.is_tensor(attributes[k][0]):
                attributes[k] = torch.stack(v, dim=0)
            elif isinstance(attributes[k][0], float):
                attributes[k] = torch.tensor(v, dtype=torch.float)
            elif isinstance(attributes[k][0], int):
                attributes[k] = torch.tensor(v, dtype=torch.long)

        if indices is not None:
            node_to_tensor_mapping = \
                torch.zeros([self.G.number_of_nodes(), ], dtype=torch.int64)
            for node_type in indices:
                # row 0 for graph index, row 1 for tensor index
                indices[node_type] = torch.tensor(indices[node_type],
                                                  dtype=torch.int64)
                node_to_tensor_mapping[indices[node_type]] = \
                    torch.arange(len(indices[node_type]), dtype=torch.int64)
            self.node_to_graph_mapping = indices
            self.node_to_tensor_mapping = node_to_tensor_mapping

        return attributes

    def _get_edge_attributes(self, key: str):
        r"""
        Similar to the `_get_node_attributes`
        """
        attributes = {}
        indices = None
        if key == "edge_feature":
            indices = {}
        for i, (_, edge_dict) in enumerate(self.G.edges.items()):
            if key in edge_dict:
                edge_type = self._get_edge_type(edge_dict)
                if edge_type not in attributes:
                    attributes[edge_type] = []
                    if indices is not None:
                        indices[edge_type] = []
                attributes[edge_type].append(edge_dict[key])
                if indices is not None:
                    # use range(0 ~ num_edges) as the graph edge indices
                    indices[edge_type].append(i)
        if len(attributes) == 0:
            return None

        for k, v in attributes.items():
            if torch.is_tensor(attributes[k][0]):
                attributes[k] = torch.stack(v, dim=0)
            elif isinstance(attributes[k][0], float):
                attributes[k] = torch.tensor(v, dtype=torch.float)
            elif isinstance(attributes[k][0], int):
                attributes[k] = torch.tensor(v, dtype=torch.long)

        if indices is not None:
            edge_to_tensor_mapping = \
                torch.zeros([self.G.number_of_edges(), ], dtype=torch.int64)
            for edge_type in indices:
                # row 0 for graph index, row 1 for tensor index
                indices[edge_type] = \
                    torch.tensor(indices[edge_type], dtype=torch.int64)
                edge_to_tensor_mapping[indices[edge_type]] = \
                    torch.arange(len(indices[edge_type]), dtype=torch.int64)
            self.edge_to_graph_mapping = indices
            self.edge_to_tensor_mapping = edge_to_tensor_mapping

        return attributes

    def _update_index(self, init: bool = False):
        r"""
        Currently store the edge_index and edge_indices for each edge_type
        """
        keys = list(self.G.nodes)
        vals = range(self.num_nodes)
        mapping = dict(zip(keys, vals))
        self.G = nx.relabel_nodes(self.G, mapping, copy=True)
        self.edge_index = \
            self._edge_to_index(list(self.G.edges(data=True)),
                                list(self.G.nodes(data=True)))

        if init:
            self.edge_label_index = self.edge_index
            self.node_label_index = {}
            for node_type in self.node_feature:
                self.node_label_index[node_type] = \
                    torch.arange(self.get_num_nodes(node_type),
                                 dtype=torch.long)

    def _edge_to_index(self, edges, nodes):
        r"""
        Make edge_index from networkx Graph Nodes and Edges.
        """
        if (len(edges) == 0):
            raise ValueError('in _edge_index, len(edges) must be '
                             'larger than 0')
        edge_index = None
        if len(edges[0]) > 2 and not isinstance(nodes[0], int) \
                and "node_type" in nodes[0][1]:
            edge_index = {}
            nodes_dict = {}
            for node in nodes:
                nodes_dict[node[0]] = node[1]['node_type']

        if len(edges[0]) > 2:
            for idx, edge in enumerate(edges):
                if isinstance(edge_index, dict):
                    edge_type = self._get_edge_type(edge[2])
                    head_type = nodes_dict[edge[0]]
                    tail_type = nodes_dict[edge[1]]
                    message_type = (head_type, edge_type, tail_type)

                    # TODO: change to support `dynamic graph` later
                    if message_type not in edge_index:
                        edge_index[message_type] = [(edge[0], edge[1])]
                    else:
                        edge_index[message_type].append((edge[0], edge[1]))

        if isinstance(edge_index, dict):
            for key in edge_index:
                edge_index[key] = torch.LongTensor(edge_index[key])

        if self.is_undirected():
            if isinstance(edge_index, dict):
                for key in edge_index:
                    edge_index[key] = torch.cat(
                        [edge_index[key], torch.flip(edge_index[key], [1])], dim=0)

        if isinstance(edge_index, dict):
            for key in edge_index:
                permute_tensor = edge_index[key].permute(1, 0)
                source_node_index = \
                    self._convert_to_tensor_index(permute_tensor[0])
                target_node_index = \
                    self._convert_to_tensor_index(permute_tensor[1])
                edge_index[key] = \
                    torch.stack([source_node_index, target_node_index])

        return edge_index

    @staticmethod
    def _is_edge_attribute(key: str) -> bool:
        r"""
        Check whether an attribute is a edge attribute.
        """
        # could be feature, label, etc.
        return 'edge' in key and 'index' not in key and 'type' not in key

    @staticmethod
    def _is_node_attribute(key: str) -> bool:
        r"""
        Check whether an attribute is a node attribute.
        """
        # could be feature, label, etc.
        return 'node' in key and 'index' not in key and 'type' not in key

    def _is_valid(self):
        r"""
        Check validity.
        """
        for key in self.keys:
            if self._is_node_attribute(key):
                num_nodes = 0
                if key != 'node_to_tensor_mapping':
                    for node_type in self[key]:
                        num_nodes += self[key][node_type].size(0)
                else:
                    num_nodes = self[key].size(0)
                assert self.num_nodes == num_nodes,\
                    'key {} is not valid'.format(key)
            if self._is_edge_attribute(key):
                num_edges = 0
                if key != 'edge_to_tensor_mapping':
                    for edge_type in self[key]:
                        num_edges += self[key][edge_type].size(0)
                else:
                    num_edges = self[key].size(0)
                assert self.num_edges == num_edges or \
                    self.num_edges * 2 == num_edges, \
                    'key {} is not valid'.format(key)

    def get_num_dims(self, key, obj_type, as_label: bool = False) -> int:
        r"""
        Returns the number of dimensions for one graph/node/edge property
        for specified types.

        Args:
            key (str): The choosing property.
            obj_type: Node or edge type.
            as_label (bool): If as_label, treat the tensor as labels.
        """
        if as_label:
            # treat as label
            if self[key] is not None and obj_type in self[key]:
                if self[key][obj_type].dtype == torch.long:
                    # classification label
                    return torch.max(self[key][obj_type]).item() + 1
                else:
                    # regression label
                    if len(self[key][obj_type].shape) == 1:
                        return 1
                    else:
                        return self.edge_label[obj_type].shape[1]
            else:
                return 0
        else:
            # treat as feature
            if self[key][obj_type] is not None:
                return self[key][obj_type].shape[1]
            else:
                return 0

    def _create_label_link_pred(self, graph, edges, nodes):
        r"""
        Create edge label and the corresponding label_index (edges) fo  r link prediction.

        Modifies the graph argument by setting the fields edge_label_i  ndex and edge_label.
        """
        graph.edge_label_index = \
            self._edge_to_index(edges, nodes)
        graph.edge_label = self._get_edge_attributes_by_key(
                edges, 'edge_label')
        graph._objective_edges = edges

    def _split_node(self, split_types: List[str], split_ratio: float):
        r"""
        Split the graph into len(split_ratio) graphs for node prediction.
        Internally this splits node indices, and the model will only compute
        loss for the embedding of
        nodes in each split graph.
        In node classification, the whole graph is observed in train/val/test
        Only split over node_label_index
        """
        if split_types is None:
            split_types = self.node_types
        if not isinstance(split_types, list):
            raise TypeError('Split_types must be string or list of string.')
        if not all([split_type in self.node_types
                   for split_type in split_types]):
            raise ValueError('Split type in split_types must exist in self.node_label_index.')
        if not all(num_node_type >= len(split_ratio)
                   for split_type, num_node_type
                   in self.get_num_nodes(split_types).items()):
            raise ValueError('In _split_node num of nodes of a specific type is smaller than the '
                             'number of splitted parts.')
        split_graphs = []
        split_offsets = {}
        split_type_index_lengths = {}
        split_type_indices = {}

        for i, split_ratio_i in enumerate(split_ratio):
            graph_new = copy.copy(self)
            node_label_index = {}
            for split_type in split_types:
                if split_type not in split_offsets:
                    split_offsets[split_type] = 0
                    split_type_index_lengths[split_type] = \
                        len(graph_new.node_label_index[split_type])
                    split_type_indices[split_type] = \
                        graph_new.node_label_index[split_type][torch.randperm(
                            split_type_index_lengths[split_type])]
                split_offset = split_offsets[split_type]
                split_type_index_length = split_type_index_lengths[split_type]
                split_type_index = split_type_indices[split_type]

                # perform `secure split` s.t. guarantees all splitted subgraph
                # of a split type contains at least one node.
                if i != len(split_ratio) - 1:
                    num_split_i = 1 + \
                        int(split_ratio_i *
                            (split_type_index_length - len(split_ratio)))
                    nodes_split_i = \
                        split_type_index[split_offset: split_offset + num_split_i]
                    split_offsets[split_type] += num_split_i
                else:
                    nodes_split_i = split_type_index[split_offset:]

                node_label_index[split_type] = nodes_split_i

            # add the non-splitted types
            for non_split_type in self.node_types:
                if (non_split_type not in split_types):
                    node_label_index[non_split_type] = self.node_label_index[non_split_type]

            graph_new.node_label_index = node_label_index
            split_graphs.append(graph_new)

        return split_graphs

    def _split_edge(self, split_types: List[tuple], split_ratio: float):
        r"""
        Split the graph into len(split_ratio) graphs for node prediction.
        Internally this splits node indices, and the model will only compute
        loss for the embedding of nodes in each split graph.
        In edge classification, the whole graph is observed in train/val/test.
        Only split over edge_label_index.
        """
        if split_types is None:
            split_types = self.message_types
        if not isinstance(split_types, list):
            raise TypeError('Split_types must be string or list of string.')
        if not all([split_type in self.message_types
                    for split_type in split_types]):
            raise ValueError('Split type in split_types must exist in self.node_label_index.')
        if not all(num_edge_type >= len(split_ratio)
                   for split_type, num_edge_type
                   in self.get_num_edges(split_types).items()):
            raise ValueError('In _split_edge num of edges of a specific type is smaller than the '
                             'number of splitted parts.')

        split_graphs = []
        split_offsets = {}
        split_type_index_lengths = {}
        split_type_indices = {}

        for i, split_ratio_i in enumerate(split_ratio):
            graph_new = copy.copy(self)
            edge_label_index = {}
            for split_type in split_types:
                if (split_type not in split_offsets):
                    split_offsets[split_type] = 0
                    split_type_index_lengths[split_type] = \
                        graph_new.edge_label_index[split_type].shape[1]
                    rand_idx_type = \
                        torch.randperm(split_type_index_lengths[split_type])
                    split_type_indices[split_type] = \
                        graph_new.edge_label_index[split_type][:, rand_idx_type]

                split_offset = split_offsets[split_type]
                split_type_index_length = split_type_index_lengths[split_type]
                split_type_index = split_type_indices[split_type]

                # perform `secure split` s.t. guarantees all splitted subgraph
                # of a split type contains at least one edge.
                if i != len(split_ratio) - 1:
                    num_split_i = 1 + \
                        int(split_ratio_i *
                            (split_type_index_length - len(split_ratio)))
                    edges_split_i = \
                        split_type_index[:, split_offset: split_offset + num_split_i]
                    split_offsets[split_type] += num_split_i
                else:
                    edges_split_i = split_type_index[:, split_offset:]

                edge_label_index[split_type] = edges_split_i

            # add the non-splitted types
            for non_split_type in self.message_types:
                if (non_split_type not in split_types):
                    edge_label_index[non_split_type] = \
                        self.edge_label_index[non_split_type]

            graph_new.edge_label_index = edge_label_index
            split_graphs.append(graph_new)

        return split_graphs

    def split_link_pred(
            self,
            split_types: List[tuple],
            split_ratio: Union[float, List[float]],
            edge_split_mode: str = "exact"):
        r"""
        Split the graph into len(split_ratio) graphs for link prediction.
        Internally this splits edge indices, and the model will only compute
        loss for the embedding of
        nodes in each split graph.
        This is only used for transductive link prediction task
        In this task, different part of graph is observed in train/val/test
        Note: this functon will be called twice,
        if during training, we further split the training graph so that
        message edges and objective edges are different
        """
        if split_types is None:
            split_types = self.message_types
        if not isinstance(split_types, list):
            raise TypeError('Split_types must be string or list of string.')
        if not all([split_type in self.message_types
                    for split_type in split_types]):
            raise ValueError('Split type in split_types must exist in self.node_label_index')
        if isinstance(split_ratio, float):
            split_ratio = [split_ratio, 1 - split_ratio]
        if len(split_ratio) < 2 or len(split_ratio) > 3:
            raise ValueError('Unrecoginzed number of splits')

        if not all(num_edge_type >= len(split_ratio)
                   for split_type, num_edge_type
                   in self.get_num_edges(split_types).items()):
            raise ValueError('in _split_edge num of edges of a specific type is smaller than the '
                             'number of splitted parts')

        split_types_all_flag = split_types == self.message_types

        if (edge_split_mode == "approximate" and not split_types_all_flag):
            warnings.warn('in _split_edge when edge_split_mode is set to be approximate '
                          'split_types does not cover all types, in this case '
                          'the splitting is not going to improve much compared '
                          'to when _split_edge when edge_split_mode is exact ')

        # Split edges by going through edges in each edge types and divide them
        # according to split_ratio.
        if edge_split_mode == 'exact':
            edges_train, edges_val, edges_test = [], [], []
            edges_split_type_dict = {}
            for edge in self.G.edges(data=True):
                edge_type = edge[2]['edge_type']
                source_node_idx = edge[0]
                target_node_idx = edge[1]
                source_node_type = self.G.nodes[source_node_idx]['node_type']
                target_node_type = self.G.nodes[target_node_idx]['node_type']
                edge_split_type = (source_node_type, edge_type, target_node_type)

                if edge_split_type not in edges_split_type_dict:
                    edges_split_type_dict[edge_split_type] = []
                edges_split_type_dict[edge_split_type].append(edge)
            for split_type in self.message_types:
                edges_split_type = edges_split_type_dict[split_type]
                edges_split_type_length = len(edges_split_type)
                random.shuffle(edges_split_type)
                if len(split_ratio) == 2:
                    if split_type in split_types:
                        num_edges_train = \
                            1 + int(split_ratio[0] * (edges_split_type_length - 2))
                        edges_train += edges_split_type[:num_edges_train]
                        edges_val += edges_split_type[num_edges_train:]
                    else:
                        edges_train += edges_split_type
                        edges_val += edges_split_type

                # perform `secure split` s.t. guarantees all splitted subgraph
                # of a split type contains at least one edge.
                elif len(split_ratio) == 3:
                    if split_type in split_types:
                        num_edges_train = \
                            1 + int(split_ratio[0] * (edges_split_type_length - 3))
                        num_edges_val = \
                            1 + int(split_ratio[1] * (edges_split_type_length - 3))

                        edges_train += edges_split_type[:num_edges_train]
                        edges_val += \
                            edges_split_type[num_edges_train:num_edges_train + num_edges_val]
                        edges_test += \
                            edges_split_type[num_edges_train + num_edges_val:]
                    else:
                        edges_train += edges_split_type
                        edges_val += edges_split_type
                        edges_test += edges_split_type

        elif edge_split_mode == "approximate":
            # if split_types do not cover all edge types in hetero_graph
            # then we need to do the filtering
            if not split_types_all_flag:
                edges_split_type = []
                edges_non_split_type = []
                for edge in self.G.edges(data=True):
                    edge_type = edge[2]['edge_type']
                    source_node_idx = edge[0]
                    target_node_idx = edge[1]
                    source_node_type = self.G.nodes[source_node_idx]['node_type']
                    target_node_type = self.G.nodes[target_node_idx]['node_type']
                    edge_split_type = (source_node_type, edge_type, target_node_type)
                    if edge_split_type in split_types:
                        edges_split_type.append(edge)
                    else:
                        edges_non_split_type.append(edge)

                random.shuffle(edges_split_type)
                edges_split_type_length = len(edges_split_type)

                # perform `secure split` s.t. guarantees all splitted subgraph
                # of a split type contains at least one edge.
                if len(split_ratio) == 2:
                    num_edges_train = \
                        1 + int(split_ratio[0] * (edges_split_type_length - 2))

                    edges_train = \
                        edges_split_type[:num_edges_train] + edges_non_split_type
                    edges_val = \
                        edges_split_type[num_edges_train:] + edges_non_split_type
                elif len(split_ratio) == 3:
                    num_edges_train = \
                        1 + int(split_ratio[0] * (edges_split_type_length - 3))
                    num_edges_val = \
                        1 + int(split_ratio[1] * (edges_split_type_length - 3))

                    edges_train = \
                        edges_split_type[:num_edges_train] + edges_non_split_type
                    edges_val = \
                        edges_split_type[num_edges_train:num_edges_train + num_edges_val] + \
                        edges_non_split_type
                    edges_test = \
                        edges_split_type[num_edges_train + num_edges_val:] + \
                        edges_non_split_type
            else:
                # if split_types cover all edge types in hetero_graph then we
                # need not do the filtering and achieve maximal optimization
                # as compared to exact split by splitting all the edges
                # regardless of edge types
                edges = list(self.G.edges(data=True))
                random.shuffle(edges)

                # perform `secure split` s.t. guarantees all splitted subgraph
                # contains at least one edge.
                if len(split_ratio) == 2:
                    num_edges_train = 1 + int(split_ratio[0] * (self.num_edges - 2))

                    edges_train = edges[:num_edges_train]
                    edges_val = edges[num_edges_train:]
                elif len(split_ratio) == 3:
                    num_edges_train = 1 + int(split_ratio[0] * (self.num_edges - 3))
                    num_edges_val = 1 + int(split_ratio[1] * (self.num_edges - 3))

                    edges_train = edges[:num_edges_train]
                    edges_val = edges[num_edges_train:num_edges_train + num_edges_val]
                    edges_test = edges[num_edges_train + num_edges_val:]

        graph_train = HeteroGraph(
            self._edge_subgraph_with_isonodes(self.G, edges_train))
        graph_val = copy.copy(graph_train)
        if len(split_ratio) == 3:
            graph_test = HeteroGraph(self._edge_subgraph_with_isonodes(
                self.G, edges_train + edges_val))

        # set objective
        self._create_label_link_pred(graph_train, edges_train, list(self.G.nodes(data=True)))
        self._create_label_link_pred(graph_val, edges_val, list(self.G.nodes(data=True)))
        if len(split_ratio) == 3:
            self._create_label_link_pred(graph_test, edges_test, list(self.G.nodes(data=True)))
            return [graph_train, graph_val, graph_test]
        else:
            return [graph_train, graph_val]

    def split(
            self,
            task: str = 'node',
            split_types: Union[str, List[str], tuple, List[tuple]] = None,
            split_ratio: List[float] = None,
            edge_split_mode: str = "exact"):
        r"""
        Split current graph object to list of graph objects.

        Args:
            task (string): One of `node`, `edge` or `link_pred`.
            split_types (list): Types splitted on. Default is `None` which will split all the types in
                specified task.
            split_ratio (array_like): Array_like ratios `[train_ratio, validation_ratio, test_ratio]`.

        Returns:
            list: A Python list of Graph objects with specified task.
        """
        if split_ratio is None:
            split_ratio = [0.8, 0.1, 0.1]
        if not isinstance(split_ratio, list):
            raise TypeError('Split ratio must be a list.')
        if (len(split_ratio) != 3):
            raise ValueError('Split ratio must contain three values.')
        if not math.isclose(sum(split_ratio), 1.0):
            raise ValueError('Split ratio must sum up to 1.')
        if not all(isinstance(split_ratio_i, float) for
                   split_ratio_i in split_ratio):
            raise TypeError('Split ratio must contain all floats.')
        if not all(split_ratio_i > 0 for split_ratio_i in split_ratio):
            raise ValueError('Split ratio must contain all positivevalues.')

        # Support for split_types passed in as string instead of list of strings
        if isinstance(split_types, str):
            split_types = [split_types]

        if task == 'node':
            return self._split_node(split_types, split_ratio)
        elif task == 'edge':
            return self._split_edge(split_types, split_ratio)
        elif task == 'link_pred':
            return self.split_link_pred(split_types, split_ratio, edge_split_mode)
        elif task == 'graph':
            raise ValueError('Graph task does not split individual graphs.')
        else:
            raise ValueError('Unknown task.')

    def _create_neg_sampling(
            self,
            negative_sampling_ratio: float,
            split_types: List[str] = None,
            resample: bool = False):
        r"""
        Create negative samples for link prediction,
        and changes the edge_label and edge_label_index accordingly (if already existed).

        Simplest link prediction has no label. It will be treated as binary classification.
        edge_label will be set to 1 for positives and 0 for negative examples.

        For link prediction that requires prediction of edge type, it will be a multi-class
        classification task.
        edge_label will be set to the (original label + 1) for positives and 0 for negative
        examples. Hence the number of prediction classes will be incremented by 1.
        In this case dataset.num_edge_labels should be called after split
        (which calls this function).

        Args:
            negative_sampling_ratio (float or int): ratio of negative sampling edges compared with the original edges.
            resample (boolean): whether should resample.
        """
        if split_types is None:
            split_types = self.message_types
        if not isinstance(split_types, list):
            raise TypeError('Split_types must be string or list of string.')
        if not all([split_type in self.message_types
                    for split_type in split_types]):
            raise ValueError('Split type in split_types must exist in self.node_label_index')

        # filter split_types
        split_types = \
            [message_type for message_type in split_types
                if message_type in self.edge_label_index]

        if resample and self._num_positive_examples is not None:
            self.edge_label_index = \
                {message_type:
                    self.edge_label_index[message_type][:, :edge_type_positive_num]
                    for message_type, edge_type_positive_num in
                    self._num_positive_examples.items()}
        num_pos_edges = \
            {message_type: edge_type_positive.shape[-1]
                for message_type, edge_type_positive in
                self.edge_label_index.items()}
        num_neg_edges = \
            {message_type: int(edge_type_num * negative_sampling_ratio)
                for message_type, edge_type_num in num_pos_edges.items()
                if message_type in split_types}

        if set(self.edge_index.keys()) == set(self.edge_label_index.keys()) and \
                all(self.edge_index[message_type].size(1) ==
                    self.edge_label_index[message_type].size(1)
                    for message_type in split_types) and \
                all(torch.sum(self.edge_index[message_type] -
                    self.edge_label_index[message_type]) == 0
                    for message_type in split_types):
            edge_index_all = \
                {message_type: edge_type_positive
                    for message_type, edge_type_positive
                    in self.edge_index.items() if message_type in split_types}
        else:
            edge_index_all = {}
            for message_type in split_types:
                edge_index_all[message_type] = \
                    torch.cat((self.edge_index[message_type],
                               self.edge_label_index[message_type]), -1)

        negative_edges = \
            self.negative_sampling(edge_index_all,
                                   self.get_num_nodes(),
                                   num_neg_edges)

        negative_label = \
            {message_type: torch.zeros(edge_type_negative, dtype=torch.long)
                for message_type, edge_type_negative in num_neg_edges.items()
                if message_type in split_types}

        if resample:
            positive_label = \
                {message_type: edge_type_positive[:num_pos_edges[message_type]]
                    for message_type, edge_type_positive in self.edge_label.items()
                    if message_type in self.edge_label_index}
        elif self.edge_label is None:
            positive_label = \
                {message_type: torch.ones(num_pos_edges[message_type], dtype=torch.long)
                    for message_type in self.edge_label_index}
        else:
            positive_label = \
                {message_type: edge_type_positive + 1
                    for message_type, edge_type_positive in self.edge_label
                    if message_type in self.edge_label_index}

        self._num_positive_examples = num_pos_edges

        for message_type in split_types:
            self.edge_label_index[message_type] = \
                torch.cat((self.edge_label_index[message_type],
                           negative_edges[message_type]), -1)

        self.edge_label = \
            {message_type:
                torch.cat((positive_label[message_type],
                           negative_label[message_type]), -1).type(torch.long)
                for message_type in split_types}

        # for message_type in self.message_types:
        for message_type in self.edge_label_index:
            if message_type not in split_types:
                self.edge_label[message_type] = positive_label[message_type]

    @staticmethod
    def negative_sampling(
            edge_index: Dict[str, torch.tensor],
            num_nodes=None,
            num_neg_samples: Dict[str, int] = None):
        r"""Samples random negative edges of a heterogeneous graph given by :attr:`edge_index`.

        Args:
            edge_index (LongTensor): The edge indices.
            num_nodes (int, optional): The number of nodes, *i.e.*
                :obj:`max_val + 1` of :attr:`edge_index`. (default: :obj:`None`)
            num_neg_samples (int, optional): The number of negative samples to
                return. If set to :obj:`None`, will try to return a negative edge
                for every positive edge. (default: :obj:`None`)
            force_undirected (bool, optional): If set to :obj:`True`, sampled
                negative edges will be undirected. (default: :obj:`False`)

        :rtype: :class:`torch.LongTensor`
        """
        num_neg_samples = \
            {message_type: min(num_neg_samples[message_type],
                               num_nodes[message_type[0]] *
                               num_nodes[message_type[2]] -
                               edge_index[message_type].size(1))
                for message_type in edge_index}

        rng = \
            {message_type:
                range(num_nodes[message_type[0]] * num_nodes[message_type[2]])
             for message_type in edge_index}
        idx = \
            {message_type:
                (edge_index[message_type][0] *
                 num_nodes[message_type[0]] +
                 edge_index[message_type]).to('cpu')
             for message_type in edge_index}

        perm = \
            {message_type: torch.tensor(random.sample(rng[message_type],
             num_neg_samples[message_type]))
                for message_type in edge_index}
        mask = \
            {message_type: torch.from_numpy(np.isin(perm[message_type],
             idx[message_type])).to(torch.bool)
                for message_type in edge_index}
        rest = \
            {message_type: mask[message_type].nonzero().view(-1)
                for message_type in edge_index}

        for message_type in edge_index:
            while rest[message_type].numel() > 0:
                tmp = torch.tensor(
                        random.sample(rng[message_type],
                                      rest[message_type].size(0)))
                mask = \
                    torch.from_numpy(
                        np.isin(tmp, idx[message_type])).to(torch.bool)
                perm[message_type][rest[message_type]] = tmp
                rest[message_type] = \
                    rest[message_type][mask.nonzero().view(-1)]

        row = \
            {message_type: perm[message_type] / num_nodes[message_type[0]]
                for message_type in perm}
        col = \
            {message_type: perm[message_type] % num_nodes[message_type[0]]
                for message_type in perm}
        neg_edge_index = \
            {message_type: torch.stack([row[message_type],
             col[message_type]], dim=0).long()
                for message_type in edge_index}

        for message_type in edge_index:
            neg_edge_index[message_type].to(neg_edge_index[message_type].device)
        return neg_edge_index

    def __cat_dim__(self, key: str, value) -> int:
        r"""
        Returns the dimension for which :obj:`value` of attribute
        :obj:`key` will get concatenated when creating batches.

        .. note::

          This method is for internal use only, and should only be overridden
          if the batch concatenation process is corrupted for a specific data
          attribute.
        """
        # `*index*` and `*face*` should be concatenated in the last dimension,
        # everything else in the first dimension.
        # filter out the node_edge_index_map, whose keys are tuples
        return -1 if isinstance(key, tuple) else 0

    def __inc__(self, key: str, value) -> int:
        r""""
        Returns the incremental count to cumulatively increase the value
        of the next attribute of :obj:`key` when creating batches.

        .. note::

          This method is for internal use only, and should only be overridden
          if the batch concatenation process is corrupted for a specific data
          attribute.
        """
        # Only `*index*` and `*face*` should be cumulatively summed up when
        # creating batches.
        # filter out the node_edge_index_map, whose keys are tuples
        if not isinstance(key, tuple):
            return 0
        node_type_start, _, node_type_end = key
        return torch.tensor([[self.get_num_nodes(node_type_start)],
                             [self.get_num_nodes(node_type_end)]])



import re 
import random
import copy
import math
import pdb
import numpy as np
import torch
import networkx as nx
from torch_geometric.utils import to_undirected
from typing import *


class Graph(object):
    r"""
    A plain python object modeling a single graph with various
    (optional) attributes:

    Args:
        G (:class:`networkx.classes.graph`): The NetworkX graph object which contains features
             and labels for the tasks.
        **kwargs: keyworded argument list with keys such as :obj:`"node_feature"`, :obj:`"node_label"`
            and corresponding attributes.
    """
    def __init__(self, G=None, **kwargs):
        self.G = G
        if G is not None:
            keys = ['node_feature', 'node_label', 'edge_feature', 'edge_label',
                    'graph_feature', 'graph_label',
                    'edge_index', 'edge_label_index', 'node_label_index']
            for key in keys:
                self[key] = None

            for key, item in kwargs.items():
                self[key] = item

            self._update_tensors(init=True)
        self._num_positive_examples = None

    @classmethod
    def _from_dict(cls, dictionary: Dict[str, torch.tensor]):
        r"""
        Creates a data object from a python dictionary.

        Args:
            dictionary (dict): Python dictionary with key (string) - value (torch.tensor) pair.

        Returns:
            :class:`deepsnap.graph.Graph`: return a new Graph object with the data from the dictionary.
        """
        graph = cls()
        for key, item in dictionary.items():
            graph[key] = item

        return graph

    def __getitem__(self, key: str):
        r"""
        Gets the data of the attribute :obj:`key`.
        """
        return getattr(self, key, None)

    def __setitem__(self, key: str, value):
        """Sets the attribute :obj:`key` to :obj:`value`."""
        setattr(self, key, value)

    @property
    def keys(self):
        r"""
        Returns all names of the graph attributes.

        Returns:
            list: List of :class:`deepsnap.graph.Graph` attributes.
        """
        # filter attributes that are not observed by users
        # (1) those with value "None"; (2) those start with '_'
        keys = [key for key in self.__dict__.keys() if self[key] is not None]
        keys = [key for key in keys if key[:1] != '_']
        return keys

    def __len__(self) -> int:
        r"""
        Returns the number of all present attributes.

        Returns:
            int: The number of all present attributes.
        """
        return len(self.keys)

    def __contains__(self, key: str) -> bool:
        r"""
        Returns :obj:`True`, if the attribute :obj:`key` is present in the
        data.
        """
        return key in self.keys

    def __iter__(self):
        r"""
        Iterates over all present attributes in the data, yielding their
        attribute names and content.
        """
        for key in sorted(self.keys):
            yield key, self[key]

    def __call__(self, *keys):
        r"""
        Iterates over all attributes :obj:`*keys` in the data, yielding
        their attribute names and content.
        If :obj:`*keys` is not given this method will iterative over all
        present attributes.
        """
        for key in sorted(self.keys) if not keys else keys:
            if key in self:
                yield key, self[key]

    def __cat_dim__(self, key: str, value) -> int:
        r"""
        Returns the dimension for which :obj:`value` of attribute
        :obj:`key` will get concatenated when creating batches.

        .. note::

            This method is for internal use only, and should only be overridden
            if the batch concatenation process is corrupted for a specific data
            attribute.
        """
        # `*index*` and `*face*` should be concatenated in the last dimension,
        # everything else in the first dimension.
        return -1 if bool(re.search('(index)', key)) else 0

    def __inc__(self, key: str, value) -> int:
        r""""
        Returns the incremental count to cumulatively increase the value
        of the next attribute of :obj:`key` when creating batches.

        .. note::

            This method is for internal use only, and should only be overridden
            if the batch concatenation process is corrupted for a specific data
            attribute.
        """
        # Only `*index*` and `*face*` should be cumulatively summed up when
        # creating batches.
        return self.num_nodes if bool(re.search('(index)', key)) else 0

    @property
    def num_nodes(self) -> int:
        r"""
        Return number of nodes in the graph.

        Returns:
            int: Number of nodes in the graph.
        """
        return self.G.number_of_nodes()

    @property
    def num_edges(self) -> int:
        r"""
        Returns number of edges in the graph.

        Returns:
            int: Number of edges.
        """
        return self.G.number_of_edges()

    @property
    def num_node_features(self) -> int:
        r"""
        Returns node feature dimension in the graph.

        Returns:
            int: Node feature dimension and `0` if there is no `node_feature`.
        """
        return self.get_num_dims('node_feature', as_label=False)

    @property
    def num_node_labels(self) -> int:
        r"""
        Returns number of the node labels in the graph.

        Returns:
            int: Number of node labels and `0` if there is no `node_label`.
        """
        return self.get_num_dims('node_label', as_label=True)

    @property
    def num_edge_features(self) -> int:
        r"""
        Returns edge feature dimension in the graph.

        Returns:
            int: Node feature dimension and `0` if there is no `edge_feature`.
        """
        return self.get_num_dims('edge_feature', as_label=False)

    @property
    def num_edge_labels(self) -> int:
        r"""
        Returns number of the edge labels in the graph.

        Returns:
            int: Number of edge labels and `0` if there is no `edge_label`.
        """
        return self.get_num_dims('edge_label', as_label=True)

    @property
    def num_graph_features(self) -> int:
        r"""
        Returns graph feature dimension in the graph.

        Returns:
            int: Graph feature dimension and `0` if there is no
            `graph_feature`.
        """
        return self.get_num_dims('graph_feature', as_label=False)

    @property
    def num_graph_labels(self) -> int:
        r"""
        Returns number of the graph labels in the graph.

        Returns:
            int: Number of graph labels and `0` if there is no `graph_label`.
        """
        return self.get_num_dims('graph_label', as_label=True)

    def get_num_dims(self, key, as_label=False) -> int:
        r"""
        Returns the number of dimensions for one graph/node/edge property.

        Args:
            as_label: if as_label, treat the tensor as labels (
        """
        if as_label:
            # treat as label
            if self[key] is not None:
                if self[key].dtype == torch.long:
                    # classification label
                    return torch.max(self[key]).item() + 1
                else:
                    # regression label
                    if len(self[key].shape) == 1 and not Graph._is_graph_attribute(key):
                        # for node/edge tasks: 1 scalar per node/edge
                        return 1
                    else:
                        return self[key].shape[-1]
            else:
                return 0
        else:
            # treat as feature
            if self[key] is not None:
                return self[key].shape[1]
            else:
                return 0

    def is_directed(self) -> bool:
        r"""
        Whether the graph is directed.

        Returns:
            bool: :obj:`True` if the graph is directed.
        """
        return self.G.is_directed()

    def is_undirected(self) -> bool:
        r"""
        Whether the graph is undirected.

        Returns:
            bool: :obj:`True` if the graph is undirected.
        """
        return not self.G.is_directed()

    def apply_tensor(self, func, *keys):
        r"""
        Applies the function :obj:`func` to all tensor attributes
        :obj:`*keys`. If :obj:`*keys` is not given, :obj:`func` is applied to
        all present attributes.

        Args:
            func (function): a function can be applied to a PyTorch tensor.
            *keys (string, optional): names of the tensor attributes that will be applied.

        Returns:
            :class:`deepsnap.graph.Graph`: Return the self :class:`deepsnap.graph.Graph`.
        """
        for key, item in self(*keys):
            if torch.is_tensor(item):
                self[key] = func(item)
            elif isinstance(self[key], dict):
                for obj_key, obj_item in self[key].items():
                    if torch.is_tensor(obj_item):
                        self[key][obj_key] = func(obj_item)
        return self

    def contiguous(self, *keys):
        r"""
        Ensures a contiguous memory layout for the attributes specified by
        :obj:`*keys`. If :obj:`*keys` is not given, all present attributes
        are ensured tohave a contiguous memory layout.

        Args:
            *keys (string, optional): tensor attributes which will be in contiguous memory layout.

        Returns:
            :class:`deepsnap.graph.Graph`: :class:`deepsnap.graph.Graph` object with specified tensor attributes in contiguous memory layout.
        """
        return self.apply_tensor(lambda x: x.contiguous(), *keys)

    def to(self, device, *keys):
        r"""
        Performs tensor dtype and/or device conversion to all attributes
        :obj:`*keys`.
        If :obj:`*keys` is not given, the conversion is applied to all present
        attributes.

        Args:
            device: Specified device name.
            *keys (string, optional): Tensor attributes which will transfer to the specified device.
        """
        return self.apply_tensor(lambda x: x.to(device), *keys)

    def clone(self):
        r"""
        Deepcopy the graph object.

        Returns:
            :class:`deepsnap.graph.Graph`: A cloned :class:`deepsnap.graph.Graph` object with deepcopying all features.
        """
        return self.__class__._from_dict({
            k: v.clone() if torch.is_tensor(v) else copy.deepcopy(v)
            for k, v in self.__dict__.items()
        })

    def _size_repr(self, value) -> List[int]:
        r"""
        Returns:
            list: A list of size of each element in value
        """
        if torch.is_tensor(value):
            return list(value.size())
        elif isinstance(value, int) or isinstance(value, float):
            return [1]
        elif isinstance(value, list) or isinstance(value, tuple):
            return [len(value)]
        else:
            return []

    def __repr__(self):
        info = ['{}={}'.format(key, self._size_repr(item)) for key, item in self]
        return '{}({})'.format(self.__class__.__name__, ', '.join(info))

    @staticmethod
    def _is_edge_attribute(key: str) -> bool:
        r"""
        Check whether an attribute is a edge attribute.
        """
        # could be feature, label, etc.
        return 'edge' in key and 'index' not in key

    @staticmethod
    def _is_node_attribute(key: str) -> bool:
        r"""
        Check whether an attribute is a node attribute.
        """
        # could be feature, label, etc.
        return 'node' in key and 'index' not in key

    @staticmethod
    def _is_graph_attribute(key: str) -> bool:
        r"""
        Check whether an attribute is a graph attribute.
        """
        return 'graph' in key and 'index' not in key

    def _is_valid(self):
        r"""
        Check validity.
        """
        for key in self.keys:
            if self._is_node_attribute(key):
                assert self.num_nodes == self[key].size(0),\
                    'key {} is not valid, num nodes must equal num nodes w/ features'.format(key)

    def _update_tensors(self, init: bool = False):
        r"""
        Update attributes and edge indices with values from the .G graph object.
        """
        self._update_attributes()
        self._update_index(init)

    def _update_attributes(self):
        r"""
        Update attributes
        """
        # node
        if (self.G.number_of_nodes() == 0):
            raise ValueError('in _update_attributes, number of nodes in Graph '
                             'G must be larger than 0')
        if (self.G.number_of_edges() == 0):
            raise ValueError('in _update_attributes, number of edges in Graph '
                             'G must be larger than 0')

        keys = next(iter(self.G.nodes(data=True)))[-1].keys()
        for key in keys:
            if key != 'node_type':
                self[key] = self._get_node_attributes(key)
        # edge
        keys = next(iter(self.G.edges(data=True)))[-1].keys()
        for key in keys:
            if key != 'edge_type':
                self[key] = self._get_edge_attributes(key)
        # graph
        keys = self.G.graph.keys()
        for key in keys:
            self[key] = self._get_graph_attributes(key)

    def _get_node_attributes(self, name: str):
        r"""
        Returns the node attributes in the graph. Multiple attributes will be stacked.

        Args:
            name(string): the name of the attributes to return.

        Returns:
            :class:`torch.tensor`: Node attributes.
        """
        # new: concat
        attributes = []
        for _, d in self.G.nodes.items():
            if name in d:
                attributes.append(d[name])
        if len(attributes) == 0:
            return None
        if torch.is_tensor(attributes[0]):
            attributes = torch.stack(attributes, dim=0)
        elif isinstance(attributes[0], float):
            attributes = torch.tensor(attributes, dtype=torch.float)
        elif isinstance(attributes[0], int):
            attributes = torch.tensor(attributes, dtype=torch.long)

        return attributes

    def _get_edge_attributes(self, name: str):
        r"""
        Returns the edge attributes in the graph. Multiple attributes will be stacked.

        Args:
            name(string): the name of the attributes to return.

        Returns:
            :class:`torch.tensor`: Edge attributes.
        """

        # new: concat
        attributes = []
        for x in self.G.edges(data=True):
            if name in x[-1]:
                attributes.append(x[-1][name])
        if len(attributes) == 0:
            return None
        if torch.is_tensor(attributes[0]):
            attributes = torch.stack(attributes, dim=0)
        elif isinstance(attributes[0], float):
            attributes = torch.tensor(attributes, dtype=torch.float)
        elif isinstance(attributes[0], int):
            attributes = torch.tensor(attributes, dtype=torch.long)
        else:
            raise ValueError('Unknown type of key {} in edge attributes.')

        if self.is_undirected():
            attributes = torch.cat([attributes, attributes], dim=0)

        return attributes

    def _get_graph_attributes(self, name: str):
        r"""
        Returns the graph attributes.

        Args:
            name(string): the name of the attributes to return.

        Returns:
            any: graph attributes with the specified name.
        """
        return self.G.graph.get(name)

    def _update_index(self, init=False):
        # relabel graphs
        keys = list(self.G.nodes)
        vals = list(range(self.num_nodes))
        if keys != vals:
            mapping = dict(zip(keys, vals))
            self.G = nx.relabel_nodes(self.G, mapping, copy=True)
        # get edges
        self.edge_index = self._edge_to_index(list(self.G.edges))
        if init:
            # init is only true when creating the variables
            # edge_label_index and node_label_index
            self.edge_label_index = self.edge_index  # by default
            self.node_label_index = \
                torch.arange(self.num_nodes, dtype=torch.long)

    def _edge_to_index(self, edges):
        r"""
        List of G.edges to torch tensor edge_index

        Only the selected edges' edge indices are extracted.
        """
        if (len(edges) == 0):
            raise ValueError('in _edge_to_index, len(edges) must be '
                             'larger than 0')
        if len(edges[0]) > 2:  # edges have features
            edges = [(edge[0], edge[1]) for edge in edges]
        edge_index = torch.LongTensor(edges)
        if self.is_undirected():
            edge_index = torch.cat(
                [edge_index, torch.flip(edge_index, [1])], dim=0)
        return edge_index.permute(1, 0)

    def _get_edge_attributes_by_key(self, edges, key: str):
        r"""
        List of G.edges to torch tensor for key, with dimension [num_edges x key_dim].

        Only the selected edges' attributes are extracted.
        """

        if (len(edges) == 0):
            raise ValueError('in _get_edge_attributes_by_key, '
                             'len(edges) must be larger than 0')
        if len(edges[0]) <= 2 or key not in edges[0][-1]:
            return None

        attributes = [edge[-1][key] for edge in edges]
        if torch.is_tensor(attributes[0]):
            attributes = torch.stack(attributes, dim=0)
        elif isinstance(attributes[0], float):
            attributes = torch.tensor(attributes, dtype=torch.float)
        elif isinstance(attributes[0], int):
            attributes = torch.tensor(attributes, dtype=torch.long)
        else:
            raise ValueError('Unknown type of key {} in edge attributes.')

        if self.is_undirected():
            attributes = torch.cat([attributes, attributes], dim=0)

        return attributes

    def _update_graphs(self, verbose=False):
        r"""
        Update the .G graph object with new attributes.
        The edges remain unchanged (edge_index should not be directly modified).

        The counter-part of update_tensors.
        """
        for key in self.keys:
            if Graph._is_node_attribute(key):
                Graph.add_node_attr(self.G, key, self[key])
            elif Graph._is_edge_attribute(key):
                # the order of edge attributes is consistent with edge index
                Graph.add_edge_attr(self.G, key, self[key])
            elif Graph._is_graph_attribute(key):
                Graph.add_graph_attr(self.G, key, self[key])
            else:
                if verbose:
                    print('Index fields: {} ignored.'.format(key))

    def apply_transform(self, transform, update_tensor: bool = True,
                        update_graph: bool = False,
                        deep_copy: bool = False, **kwargs):
        r"""
        Apply transform function to current graph object.

        Note that when the backend graph object (e.g. networkx object) is changed in the 
        transform function, the argument update_tensor is recommended, to update the tensor
        representation to be in sync with the transformed graph.
        Similarly, update_graph is recommended when the transform function makes change to
        the tensor objects.

        However, the transform function should not make changes to both the backend graph
        object and the tensors simultaneously. Otherwise there might exist inconsistency
        between the transformed graph and tensors.
        Also note that update_tensor and update_graph cannot be true at the same time.

        Args:
            transform (fuction): in the format of :obj:`transform(deepsnap.graph.Graph, **kwargs)`.
                The function needs to either return deepsnap.graph.Graph (the transformed graph
                object), or the transformed internal .G object (networkx).
                If returning .G object, all corresponding tensors will be updated.
            update_tensor (boolean): if nx graph has changed,
                use nx graph to update tensor attributes.
            update_graph: (boolean): if tensor attributes has changed,
                use attributes to update nx graph.
            deep_copy (boolean): True if a new copy of graph_object is needed.
                In this case, the transform function needs to either return a graph object,
                Important: when returning Graph object in transform function, user should decide
                whether the tensor values of the graph is to be copied (deep copy).
            **kwargs (any): additional args for the transform function.
        
        Note:
            This function different from the function :obj:`apply_tensor`.
        """
        assert not (update_tensor and update_graph)
        graph_obj = copy.deepcopy(self) if deep_copy else self
        return_graph = transform(graph_obj, **kwargs)

        if isinstance(return_graph, self.G.__class__):
            return_graph = Graph(return_graph)
        elif isinstance(return_graph, self.__class__):
            return_graph = return_graph
        elif return_graph is None:
            # no return value; assumes in-place transform of the graph object
            return_graph = graph_obj
        else:
            raise TypeError('Transform function returns a value of unknown type ({})'.format(
                return_graph.__class__))
        if update_graph:
            return_graph._update_graphs()
        if update_tensor:
            return_graph._update_tensors()
        return return_graph

    def split(self, task: str = 'node', split_ratio: List[float] = None):
        r"""
        Split current graph object to list of graph objects.

        Args:
            task (string): one of `node`, `edge` or `link_pred`.
            split_ratio (array_like): array_like ratios `[train_ratio, validation_ratio, test_ratio]`.

        Returns:
            list: A Python list of :class:`deepsnap.graph.Graph` objects with specified task.
        """
        if split_ratio is None:
            split_ratio = [0.8, 0.1, 0.1]
        if not isinstance(split_ratio, list):
            raise TypeError('split ratio must be a list.')
        if (len(split_ratio) > 3):
            raise ValueError('split ratio must contain leq three values')
        if not math.isclose(sum(split_ratio), 1.0):
            raise ValueError('split ratio must sum up to 1.')
        if not all(isinstance(split_ratio_i, float) for
                   split_ratio_i in split_ratio):
            raise TypeError('split ratio must contain all floats')
        if not all(split_ratio_i > 0 for split_ratio_i in split_ratio):
            raise ValueError('split ratio must contain all positivevalues.')

        if task == 'node':
            return self._split_node(split_ratio)
        elif task == 'edge':
            return self._split_edge(split_ratio)
        elif task == 'link_pred':
            return self.split_link_pred(split_ratio)
        elif task == 'graph':
            raise ValueError('Graph task does not split individual graphs.')
        else:
            raise ValueError('Unknown task.')

    def _split_node(self, split_ratio: float):
        r"""
        Split the graph into len(split_ratio) graphs for node prediction.
        Internally this splits node indices, and the model will only compute
        loss for the embedding of
        nodes in each split graph.
        In node classification, the whole graph is observed in train/val/test
        Only split over node_label_index
        """
        if (self.num_nodes < len(split_ratio)):
            raise ValueError('in _split_node num of nodes are smaller than'
                             'number of splitted parts')

        split_graphs = []
        shuffled_node_indices = torch.randperm(self.num_nodes)
        split_offset = 0

        # perform `secure split` s.t. guarantees all splitted subgraph
        # contains at least one node.
        for i, split_ratio_i in enumerate(split_ratio):
            if i != len(split_ratio) - 1:
                num_split_i = \
                    1 + int(split_ratio_i * (self.num_nodes - len(split_ratio)))
                nodes_split_i = shuffled_node_indices[
                                split_offset: split_offset + num_split_i]
                split_offset += num_split_i
            else:
                nodes_split_i = shuffled_node_indices[split_offset:]
            # shallow copy all attributes
            graph_new = copy.copy(self)
            graph_new.node_label_index = nodes_split_i
            split_graphs.append(graph_new)
        return split_graphs

    def _split_edge(self, split_ratio: float):
        r"""
        Split the graph into len(split_ratio) graphs for node prediction.
        Internally this splits node indices, and the model will only compute
        loss for the embedding of nodes in each split graph.
        In edge classification, the whole graph is observed in train/val/test.
        Only split over edge_label_index.
        """
        if (self.num_edges < len(split_ratio)):
            raise ValueError('in _split_node num of edges are smaller than'
                             'number of splitted parts')

        split_graphs = []
        edges = list(self.G.edges())
        random.shuffle(edges)
        split_offset = 0

        # perform `secure split` s.t. guarantees all splitted subgraph
        # contains at least one edge.
        for i, split_ratio_i in enumerate(split_ratio):
            if i != len(split_ratio) - 1:
                num_split_i = \
                    1 + int(split_ratio_i * (self.num_edges - len(split_ratio)))
                edges_split_i = edges[split_offset: split_offset + num_split_i]
                split_offset += num_split_i
            else:
                edges_split_i = edges[split_offset:]
            # shallow copy all attributes
            graph_new = copy.copy(self)
            graph_new.edge_label_index = self._edge_to_index(edges_split_i)
            split_graphs.append(graph_new)
        return split_graphs

    def split_link_pred(self, split_ratio: Union[float, List[float]]):
        r"""
        Split the graph into len(split_ratio) graphs for link prediction.
        Internally this splits edge indices, and the model will only compute
        loss for the embedding of
        nodes in each split graph.
        This is only used for transductive link prediction task
        In this task, different part of graph is observed in train/val/test
        Note: this functon will be called twice,
        if during training, we further split the training graph so that
        message edges and objective edges are different
        """
        if isinstance(split_ratio, float):
            split_ratio = [split_ratio, 1 - split_ratio]
        if len(split_ratio) < 2 or len(split_ratio) > 3:
            raise ValueError('Unrecoginzed number of splits')
        if len(split_ratio) == 2:
            if (self.num_edges < 2):
                raise ValueError('in split_link_pred num of edges are'
                                 'smaller than number of splitted parts')
        if len(split_ratio) == 3:
            if (self.num_edges < 3):
                raise ValueError('in split_link_pred num of edges are'
                                 'smaller than number of splitted parts')

        edges = list(self.G.edges(data=True))
        random.shuffle(edges)

        # perform `secure split` s.t. guarantees all splitted subgraph
        # contains at least one edge.
        if len(split_ratio) == 2:
            num_edges_train = 1 + int(split_ratio[0] * (self.num_edges - 2))

            edges_train = edges[:num_edges_train]
            edges_val = edges[num_edges_train:]
        elif len(split_ratio) == 3:
            num_edges_train = 1 + int(split_ratio[0] * (self.num_edges - 3))
            num_edges_val = 1 + int(split_ratio[1] * (self.num_edges - 3))

            edges_train = edges[:num_edges_train]
            edges_val = edges[num_edges_train:num_edges_train + num_edges_val]
            edges_test = edges[num_edges_train + num_edges_val:]

        graph_train = Graph(
            self._edge_subgraph_with_isonodes(self.G, edges_train))
        graph_val = copy.copy(graph_train)
        if len(split_ratio) == 3:
            graph_test = Graph(self._edge_subgraph_with_isonodes(
                self.G, edges_train + edges_val))

        # set objective
        self._create_label_link_pred(graph_train, edges_train)
        self._create_label_link_pred(graph_val, edges_val)
        if len(split_ratio) == 3:
            self._create_label_link_pred(graph_test, edges_test)
            return [graph_train, graph_val, graph_test]
        else:
            return [graph_train, graph_val]

    def _edge_subgraph_with_isonodes(self, G, edges):
        r"""
        Generate a new networkx graph with same nodes and their attributes.

        Preserves all nodes and a subset of edges. Nodes that are not connected
        by any of the edges will be isolated nodes instead of being removed.

        Note:
            edges should be list(G_i.edges(data=True))
        """
        G_new = G.__class__()
        G_new.add_nodes_from(G.nodes(data=True))
        G_new.add_edges_from(edges)
        return G_new

    def resample_disjoint(self, message_ratio):
        r""" Resample disjoint edge split of message passing and objective links.

        Note that if apply_transform (on the message passing graph)
        was used before this resampling, it needs to be
        re-applied, after resampling, to update some of the edges that were in objectives.
        """
        if not hasattr(self, '_objective_edges'):
            raise ValueError('No disjoint edge split was performed.')
        # combine into 1 graph
        self.G.add_edges_from(self._objective_edges)
        return self.split_link_pred(message_ratio)[1]

    def _create_label_link_pred(self, graph, edges):
        r"""
        Create edge label and the corresponding label_index (edges) for link prediction.

        Modifies the graph argument by setting the fields edge_label_index and edge_label.
        """

        graph.edge_label_index = self._edge_to_index(edges)
        graph.edge_label = self._get_edge_attributes_by_key(
                edges, 'edge_label')
        # keep a copy of original edges (and their attributes)
        # for resampling the disjoint split (message passing and objective links)
        graph._objective_edges = edges

    def _create_neg_sampling(self, negative_sampling_ratio: float, resample: bool = False):
        r"""
        Create negative samples for link prediction,
        and changes the edge_label and edge_label_index accordingly (if already existed).

        Simplest link prediction has no label. It will be treated as binary classification.
        edge_label will be set to 1 for positives and 0 for negative examples.

        For link prediction that requires prediction of edge type, it will be a multi-class
        classification task.
        edge_label will be set to the (original label + 1) for positives and 0 for negative
        examples. Hence the number of prediction classes will be incremented by 1.
        In this case dataset.num_edge_labels should be called after split
        (which calls this function).

        Args:
            negative_sampling_ratio (float or int): ratio of negative sampling edges compared with the original edges.
            resample (boolean): whether should resample.
        """
        if resample and self._num_positive_examples is not None:
            # remove previous negative samples first
            # if self._num_positive_examples is None then no previous sampling was done
            self.edge_label_index = self.edge_label_index[:, :self._num_positive_examples]
        num_pos_edges = self.edge_label_index.shape[-1]
        num_neg_edges = int(num_pos_edges * negative_sampling_ratio)

        if self.edge_index.size() == self.edge_label_index.size() and \
                (torch.sum(self.edge_index - self.edge_label_index) == 0):
            # (train in 'all' mode)
            edge_index_all = self.edge_index
        else:
            edge_index_all = torch.cat((self.edge_index, self.edge_label_index), -1)

        if (len(edge_index_all) > 0):
            negative_edges = \
                self.negative_sampling(edge_index_all, self.num_nodes, num_neg_edges)
        else:
            return torch.LongTensor([])

        # label for negative edges is 0
        negative_label = torch.zeros(num_neg_edges, dtype=torch.long)
        # positive edges
        if resample:
            # when resampling, get the positive portion of labels
            positive_label = self.edge_label[:num_pos_edges]
        elif self.edge_label is None:
            # if label is not yet specified, use all ones for positives
            positive_label = torch.ones(num_pos_edges, dtype=torch.long)
        else:
            # reserve class 0 for negatives; increment other edge labels
            positive_label = self.edge_label + 1
        self._num_positive_examples = num_pos_edges
        # append to edge_label_index
        self.edge_label_index = torch.cat(
            (self.edge_label_index, negative_edges), -1)
        self.edge_label = torch.cat((positive_label, negative_label), -1).type(torch.long)

    @staticmethod
    def add_node_attr(G, attr_name: str, node_attr):
        r"""
        Add node attribute into a NetworkX graph. Assumes that the
        `node_attr` ordering is the same as the node ordering in `G`.

        Args:
            G (NetworkX Graph): a NetworkX graph.
            attr_name (string): Name of the node
                attribute to set.
            node_attr (array_like): node attributes.
        """
        attr_dict = dict(zip(range(G.number_of_nodes()), node_attr))
        nx.set_node_attributes(G, attr_dict, name=attr_name)

    @staticmethod
    def add_edge_attr(G, attr_name: str, edge_attr):
        r"""
        Add edge attribute into a NetworkX graph.

        Args:
            G (NetworkX Graph): a NetworkX graph.
            attr_name (string): Name of the edge
                attribute to set.
            edge_attr (array_like): edge attributes.
        """
        # TODO: parallel?
        edge_list = list(G.edges())
        attr_dict = dict(zip(edge_list, edge_attr))
        nx.set_edge_attributes(G, attr_dict, name=attr_name)

    @staticmethod
    def add_graph_attr(G, attr_name: str, graph_attr):
        r"""
        Add graph attribute into a NetworkX graph.

        Args:
            G (NetworkX Graph): a NetworkX graph.
            attr_name (string): Name of the graph attribute to set.
            graph_attr (scalar or array_like): graph attributes.
        """
        G.graph[attr_name] = graph_attr

    @staticmethod
    def pyg_to_graph(data, verbose: bool = False, fixed_split: bool = False):
        r"""
        Converts Pytorch Geometric data to a Graph object.

        Args:
            data (:class:`torch_geometric.data`): a Pytorch Geometric data.
            verbose: if print verbose warning
            fixed_split: if load fixed data split from PyG dataset

        Returns:
            :class:`deepsnap.graph.Graph`: A new DeepSNAP :class:`deepsnap.graph.Graph` object.
        """
        G = nx.Graph()
        G.add_nodes_from(range(data.num_nodes))
        G.add_edges_from(data.edge_index.t().tolist())

        # all fields in PyG Data object
        kwargs = {}
        kwargs['node_feature'] = data.x if 'x' in data.keys else None
        kwargs['edge_feature'] = data.edge_attr if 'edge_attr' in data.keys \
            else None
        kwargs['node_label'], kwargs['edge_label'] = None, None
        kwargs['graph_feature'], kwargs['graph_label'] = None, None
        if kwargs['node_feature'] is not None and \
                data.y.size(0) == kwargs['node_feature'].size(0):
            kwargs['node_label'] = data.y
        elif kwargs['edge_feature'] is not None and \
                data.y.size(0) == kwargs['edge_feature'].size(0):
            kwargs['edge_label'] = data.y
        else:
            kwargs['graph_label'] = data.y

        # include other arguments that are in the kwargs of pyg data object
        keys_processed = ['x', 'y', 'edge_index', 'edge_attr']
        for key in data.keys:
            if key not in keys_processed:
                kwargs[key] = data[key]

        # we assume that edge-related and node-related features are defined
        # the same as in Graph._is_edge_attribute and Graph._is_node_attribute
        for key, value in kwargs.items():
            if value is None:
                continue
            if Graph._is_node_attribute(key):
                Graph.add_node_attr(G, key, value)
            elif Graph._is_edge_attribute(key):
                # the order of edge attributes is consistent with edge index
                Graph.add_edge_attr(G, key, value)
            elif Graph._is_graph_attribute(key):
                Graph.add_graph_attr(G, key, value)
            else:
                if verbose:
                    print('Index fields: {} ignored.'.format(key))
        if fixed_split:
            masks = ['train_mask', 'val_mask', 'test_mask']
            graph = Graph(G)
            graphs = []
            for mask in masks:
                if mask in kwargs:
                    graph_new = copy.copy(graph)
                    graph_new.node_label_index = torch.nonzero(data[mask]).squeeze()
                    graphs.append(graph_new)
            return graphs
        else:
            return Graph(G)

    @staticmethod
    def raw_to_graph(data):
        r"""
        Write other methods for user to import their own data format and
        make sure all attributes of G are scalar/torch.tensor. ``Not implemented``.
        """
        raise NotImplementedError

    @staticmethod
    def negative_sampling(edge_index, num_nodes=None, num_neg_samples=None):
        r"""Samples random negative edges of a graph given by :attr:`edge_index`.

        Args:
            edge_index (:class:`torch.LongTensor`): The edge indices.
            num_nodes (int, optional): The number of nodes, *i.e.*
                :obj:`max_val + 1` of :attr:`edge_index`. (default: :obj:`None`)
            num_neg_samples (int, optional): The number of negative samples to
                return. If set to :obj:`None`, will try to return a negative edge
                for every positive edge. (default: :obj:`None`)
            force_undirected (bool, optional): If set to :obj:`True`, sampled
                negative edges will be undirected. (default: :obj:`False`)

        :rtype: :class:`torch.LongTensor`
        """
        num_neg_samples = min(num_neg_samples,
                              num_nodes * num_nodes - edge_index.size(1))

        rng = range(num_nodes**2)
        # idx = N * i + j
        idx = (edge_index[0] * num_nodes + edge_index[1]).to('cpu')

        perm = torch.tensor(random.sample(rng, num_neg_samples))
        mask = torch.from_numpy(np.isin(perm, idx)).to(torch.bool)
        rest = mask.nonzero().view(-1)
        while rest.numel() > 0:  # pragma: no cover
            tmp = torch.tensor(random.sample(rng, rest.size(0)))
            mask = torch.from_numpy(np.isin(tmp, idx)).to(torch.bool)
            perm[rest] = tmp
            rest = rest[mask.nonzero().view(-1)]

        row = perm / num_nodes
        col = perm % num_nodes
        neg_edge_index = torch.stack([row, col], dim=0).long()

        return neg_edge_index.to(edge_index.device)



import torch
import torch
from torch._six import container_abcs
from torch_geometric.nn.inits import reset
import torch_geometric.nn as pyg_nn
import torch_geometric.utils as pyg_utils
import torch.nn as nn

class HeteroSAGEConv(pyg_nn.MessagePassing):
    r"""The heterogeneous compitable GraphSAGE operator is derived from the `"Inductive Representation 
    Learning on Large Graphs" <https://arxiv.org/abs/1706.02216>`_, `"Modeling polypharmacy side 
    effects with graph convolutional networks" <https://arxiv.org/abs/1802.00543>`_ and `"Modeling 
    Relational Data with Graph Convolutional Networks" <https://arxiv.org/abs/1703.06103>`_ papers.

    Args:
        in_channels_neigh (int): The input dimension of the end node type.
        out_channels (int): The dimension of the output.
        in_channels_self (int): The input dimension of the start node type.
            Default is `None` where the `in_channels_self` is equal to `in_channels_neigh`.
    """
    def __init__(self, in_channels_neigh, out_channels, in_channels_self=None):
        super(HeteroSAGEConv, self).__init__(aggr='add')
        self.in_channels_neigh = in_channels_neigh
        if in_channels_self is None:
            self.in_channels_self = in_channels_neigh
        else:
            self.in_channels_self = in_channels_self
        self.out_channels = out_channels
        self.lin_neigh = nn.Linear(in_channels_neigh, out_channels)
        self.lin_self = nn.Linear(in_channels_self, out_channels)
        self.lin_update = nn.Linear(out_channels * 2, out_channels)
        
    def forward(self, node_feature_neigh, node_feature_self, 
                edge_index, edge_weight=None, size=None,
                res_n_id=None):
        """"""
        edge_index, _ = pyg_utils.remove_self_loops(edge_index)
        return self.propagate(edge_index, size=size, 
                              node_feature_neigh=node_feature_neigh, node_feature_self=node_feature_self,
                              edge_weight=edge_weight, res_n_id=res_n_id)
      
    def message(self, node_feature_neigh_j, node_feature_self_i, edge_weight):
        """"""
        return node_feature_neigh_j
      
    def update(self, aggr_out, node_feature_self, res_n_id):
        """"""
        aggr_out = self.lin_neigh(aggr_out)
        node_feature_self = self.lin_self(node_feature_self)
        aggr_out = torch.cat([aggr_out, node_feature_self], dim=-1)
        aggr_out = self.lin_update(aggr_out)
        return aggr_out
      
    def __repr__(self):
        return '{}(neigh: {}, self: {}, out: {})'.format(self.__class__.__name__, self.in_channels_neigh, 
                                       self.in_channels_self, self.out_channels)

class HeteroConv(torch.nn.Module):
    r"""A "wrapper" layer designed for heterogeneous graph layers. It takes a 
    heterogeneous graph layer, such as :class:`deepsnap.hetero_gnn.HeteroSAGEConv`, at the initializing stage.
    """
    def __init__(self, convs, aggr='add', parallelize=False):
        super(HeteroConv, self).__init__()

        assert isinstance(convs, container_abcs.Mapping)
        self.convs = convs
        self.modules = torch.nn.ModuleList(convs.values())

        assert aggr in ['add', 'mean', 'max', 'mul', 'concat', None]
        self.aggr = aggr

        if parallelize and torch.cuda.is_available():
            self.streams = {key: torch.cuda.Stream() for key in convs.keys()}
        else:
            self.streams = None

    def reset_parameters(self):
        for conv in self.convs.values():
            reset(conv)

    def forward(self, node_features, edge_indices, edge_features=None):
        r"""The forward function for `HeteroConv`.

        Args:
            node_features (dict): A dictionary each key is node type and the corresponding 
                value is a node feature tensor.
            edge_indices (dict): A dictionary each key is message type and the corresponding
                value is an edge index tensor.
            edge_features (dict): A dictionary each key is edge type and the corresponding
                value is an edge feature tensor. Default is `None`.
        """
        if self.streams is not None and graph.not_in_gpu():
            raise RuntimeError('Cannot parallelize on non-gpu graphs')
        
        # node embedding computed from each message type
        message_type_emb = {}
        for message_key, message_type in edge_indices.items():
            if message_key not in self.convs:
                continue
            neigh_type, edge_type, self_type = message_key
            node_feature_neigh = node_features[neigh_type]
            node_feature_self = node_features[self_type]
            if edge_features != None:
                edge_feature = edge_features[edge_type]
            edge_index = edge_indices[message_key]

            # Perform message passing.
            if self.streams is not None:
                with torch.cuda.stream(self.streams[message_key]):
                    message_type_emb[message_key] = self.convs[message_key](node_feature_neigh, node_feature_self, edge_index)
            else:
                message_type_emb[message_key] = self.convs[message_key](node_feature_neigh, node_feature_self, edge_index)

        if self.streams is not None:
            torch.cuda.synchronize()
            
        # aggregate node embeddings from different message types into 1 node embedding for each node
        node_emb = {tail: [] for _, _, tail in message_type_emb.keys()}

        for (_, _, tail), item in message_type_emb.items():
            node_emb[tail].append(item)

        # Aggregate multiple embeddings with the same tail.
        for node_type, embs in node_emb.items():
            if len(embs) == 1:
                node_emb[node_type] = embs[0]
            else:
                node_emb[node_type] = self.aggregate(embs)

        return node_emb

    def aggregate(self, xs):
        r"""The aggregation for each node type. Currently support `concat`, `add`,
        `mean`, `max` and `mul`.
        """
        if self.aggr == 'concat':
            return torch.cat(xs, dim=-1)

        x = torch.stack(xs, dim=-1)
        if self.aggr == 'add':
            return x.sum(dim=-1)
        elif self.aggr == 'mean':
            return x.mean(dim=-1)
        elif self.aggr == 'max':
            return x.max(dim=-1)[0]
        elif self.aggr == 'mul':
            return x.prod(dim=-1)[0]

def forward_op(x, func, **kwargs):
    r"""A helper function for the heterogeneous operations. Given a dictionary input,
    it will return a dictionary with the same keys and the values applied by the 
    `func` with specified parameters.

    Args:
        x (dict): A dictionary that the value of each item will be applied by the `func`.
        func (:class:`function`): The function will be applied to each value in the dictionary.
        **kwargs: Parameters that will be passed into the `func`.
    """
    if not isinstance(x, dict):
        raise ValueError('The input x should be a dictionary')
    for key in x:
        x[key] = func(x[key], **kwargs)
    return x

def loss_op(pred, y, label_index, loss_func, **kwargs):
    r"""A helper function for the heterogeneous loss operations. 

    Args:
        pred (dict): A dictionary of predictions.
        y (dict): A dictionary of labels.
        label_index (dict): A dictionary of indicies that the loss will be computed on.
            Each value should be a Pytorch long tensor.
        loss_func (:class:`function`): The loss function.
        **kwargs: Parameters that will be passed into the `loss_func`.
    """
    loss = 0
    for key in pred:
        idx = label_index[key]
        loss += loss_func(pred[key][idx], y[key][idx])
    return loss



import copy
import math
import random
import networkx as nx
import numpy as np
from typing import *
import torch
from deepsnap.graph import Graph
from deepsnap.hetero_graph import HeteroGraph
import pdb

class Generator(object):
    r"""
    Abstract class of on the fly generator used in dataset.
    It generates graphs on the fly to be fed into the model.
    """
    def __init__(self, sizes, size_prob=None, dataset_len=0):
        self.sizes = sizes
        if sizes is not None:
            if size_prob is None:
                self.size_prob = np.ones(len(sizes)) / len(sizes)
            else:
                self.size_prob = size_prob
        # by default length of generator dataset is 0
        self._len = dataset_len

    def __len__(self):
        return self._len

    def set_len(self, dataset_len):
        self._len = dataset_len

    def _get_size(self, size=None):
        if size is None:
            return np.random.choice(
                    self.sizes, size=1, replace=True, p=self.size_prob)[0]
        else:
            return size

    @property
    def num_node_labels(self):
        return 0

    @property
    def num_nodes(self):
        return 0

    @property
    def num_edge_labels(self):
        return 0

    @property
    def num_edges(self):
        return 0

    @property
    def num_graph_labels(self):
        return 0

    def generate(self):
        r"""
        Overwrite in subclass. Generates and returns a Graph object
        """
        return Graph(nx.Graph())


class EnsembleGenerator(Generator):
    def __init__(self, generators, gen_prob=None, dataset_len=0):
        r"""
        A generator that is an ensemble of many generators.

        Args:
            generators: A list of Generators.
            prob: A list with the same length as generators.
                Specifies the probability of sampling from each generator.
                If None, uniformly sample a generator.
        """
        super(EnsembleGenerator, self).__init__(None, dataset_len=dataset_len)
        if gen_prob is None:
            self.gen_prob = np.ones(len(generators)) / len(generators)
        else:
            self.gen_prob = gen_prob
        self.generators = generators

    @property
    def num_node_labels(self):
        r"""
        Returns number of the node labels in the generated graphs.

        Returns:
            int: The number of node labels.
        """
        return max([gen.num_node_labels for gen in self.generators])

    @property
    def num_nodes(self):
        r"""
        Returns number of the nodes in each generated graphs.

        Returns:
            list: List of the number of nodes.
        """
        return [gen.num_nodes for gen in self.generators]

    @property
    def num_edge_labels(self):
        r"""
        Returns number of the edge labels in the generated graphs.

        Returns:
            int: The number of edge labels.
        """
        return max([gen.num_edge_labels for gen in self.generators])

    @property
    def num_edges(self):
        r"""
        Returns number of the edges in each generated graphs.

        Returns:
            list: List of the number of edges.
        """
        return [gen.num_edges for gen in self.generators]

    @property
    def num_graph_labels(self):
        r"""
        Returns number of the graph labels in the generated graphs.

        Returns:
            int: The number of graph labels.
        """
        return max([gen.num_graph_labels for gen in self.generators])

    def generate(self, **kwargs):
        r"""
        Generate a list of graphs.

        Returns:
            list: Generated list of :class:`deepsnap.graph.Graph` objects.
        """
        gen = np.random.choice(self.generators, 1, p=self.gen_prob)[0]
        return gen.generate(**kwargs)


class GraphDataset(object):
    r"""
        A plain python object modeling a list of Graph with various
        (optional) attributes.

        Args:
            graphs (list): A list of Graph.
            task (str): Task this GraphDataset is used for
                (task = 'node' or 'edge' or 'link_pred' or 'graph').
            edge_negative_sampling_ratio (float): The number of negative samples compared
                to that of positive data.
            edge_message_ratio (float): The number of training edge objectives
                compared to that of message-passing edges.
            edge_train_mode (str): Whether to use (edge_train_mode = 'all':
                training edge objectives are the same as the message-passing edges;
                or 'disjoint': training edge objectives are different from message-passing edges;
                or 'train_only': training edge objectives are always the training set edges).
            minimum_node_per_graph (int): If the number of nodes of a graph is smaller than this,
                that graph will be filtered out.
            generator (:class:`deepsnap.dataset.Generator`): The dataset can be on-the-fly-generated.
                When using on the fly generator, the graphs = [] or None, and
                a generator(Generator) is provided, with an overwritten
                generate() method.
        """
    def __init__(self, graphs, task: str = 'node',
                 edge_negative_sampling_ratio: float = 1,
                 edge_message_ratio: float = 0.8,
                 edge_train_mode: str = 'all',
                 edge_split_mode: str = 'exact',
                 minimum_node_per_graph: int = 5,
                 generator=None):

        if graphs is not None:
            # make sure graphs is a list
            if not isinstance(graphs, list):
                graphs = [graphs]

            # support user' input a list of nx.Graph instead of Graph
            for i, graph in enumerate(graphs):
                if isinstance(graph, nx.Graph):
                    graphs[i] = Graph(graph)

        # validity check for `task`
        if task not in ['node', 'edge', 'link_pred', 'graph']:
            raise ValueError("`task` must be one of 'node', 'edge', 'link_pred' or 'graph'")

        # validity check for `edge_train_mode`
        if edge_train_mode not in ['all', 'disjoint']:
            raise ValueError("`edge_train_mode` must be 'all' or 'disjoint'")

        # validity check for `edge_split_mode`
        if edge_split_mode not in ['exact', 'approximate']:
            raise ValueError("`edge_split_mode` must be 'exact' or 'approximate'")

        # parameter initialization
        self.graphs = graphs
        self.task = task
        self.edge_message_ratio = edge_message_ratio
        self.edge_negative_sampling_ratio = edge_negative_sampling_ratio
        self.edge_train_mode = edge_train_mode
        self.edge_split_mode = edge_split_mode
        self.minimum_node_per_graph = minimum_node_per_graph
        self._resample_negatives = False
        self._split_types = None

        # graphs preprocessing
        if graphs is None or len(graphs) == 0:
            if generator is None:
                raise ValueError('Graphs are None')
            else:
                # on-the-fly dataset
                self.generator = generator
                self.graphs = None
                self.otf_device = None
        elif generator is not None:
            raise ValueError('Both graphs and on-the-fly generator are '
                             'provided (only one should be provided.')
        else:
            # by default on-the-fly generator is not used.
            # when generator is not provide
            self.generator = None

            # filter graphs that are too small
            self.graphs = [graph for graph in self.graphs if
                           graph.num_nodes >= self.minimum_node_per_graph]

        self._reset_cache()

    def __len__(self) -> int:
        r"""
        Returns:
            int: The number of graph in graphs.
        """
        if self.graphs is None:
            return len(self.generator)
        else:
            return len(self.graphs)

    @property
    def num_node_features(self) -> int:
        r"""
        Returns node feature dimension in the graph.

        Returns:
            int: The number of features per node in the dataset.
        """
        return self._graph_example.num_node_features

    @property
    def num_node_labels(self) -> int:
        r"""
        Returns node feature dimension in the graph.

        Returns:
            int: The number of labels per node in the dataset.
        """
        if self._num_node_labels is None:
            if self.graphs is None:
                self._num_node_labels = self.generator.num_node_labels
            else:
                self._num_node_labels = \
                    max([graph.num_node_labels for graph in self.graphs])
        return self._num_node_labels

    @property
    def num_nodes(self) -> List[int]:
        r"""
        Return number of nodes in graph list

        Returns:
            list: A list of number of nodes for each graph in graph list
        """
        if self._num_nodes is None:
            if self.graphs is None:
                self._num_nodes = self.generator.num_nodes
            else:
                self._num_nodes = \
                    [graph.num_nodes for graph in self.graphs]
        return self._num_nodes

    @property
    def num_edge_features(self) -> int:
        r"""
        Returns edge feature dimension in the graph.

        Returns:
            int: The number of features per edge in the dataset.
        """
        return self._graph_example.num_edge_features

    @property
    def num_edge_labels(self) -> int:
        r"""
        Returns edge feature dimension in the graph.

        Returns:
            int: The number of labels per edge in the dataset.
        """
        if self._num_edge_labels is None:
            if self.graphs is None:
                self._num_edge_labels = self.generator.num_edge_labels
            else:
                self._num_edge_labels = \
                    max([graph.num_edge_labels for graph in self.graphs])
        return self._num_edge_labels

    @property
    def num_edges(self) -> List[int]:
        r"""
        Return number of nodes in graph list

        Returns:
            list: A list of number of nodes for each graph in graph list
        """
        if self._num_edges is None:
            if self.graphs is None:
                self._num_edges = self.generator.num_edges
            else:
                self._num_edges = \
                    [graph.num_edges for graph in self.graphs]
        return self._num_edges

    @property
    def num_graph_features(self) -> int:
        r"""
        Returns graph feature dimension in the graph.

        Returns:
            int: The number of features per graph in the dataset.
        """
        return self._graph_example.num_graph_features

    @property
    def num_graph_labels(self) -> int:
        r"""
        Returns graph feature dimension in the graph.

        Returns:
            int: The number of labels per graph in the dataset.
        """
        if self._num_graph_labels is None:
            if self.graphs is None:
                self._num_graph_labels = self.generator.num_graph_labels
            else:
                self._num_graph_labels = \
                    max([graph.num_graph_labels for graph in self.graphs])
        return self._num_graph_labels

    @property
    def num_labels(self) -> int:
        r"""
        General wrapper that returns the number of labels depending on the task.

        Returns:
            int: The number of labels, depending on the task
        """
        if self.task == 'node':
            return self.num_node_labels
        elif self.task == 'edge' or self.task == 'link_pred':
            return self.num_edge_labels
        elif self.task == 'graph':
            return self.num_graph_labels
        else:
            raise ValueError('Task {} not supported'.format(self.task))

    def num_dims_dict(self):
        r"""
        Dimensions for all fields.

        Returns:
            dict: Name of the property to the dimension.
                e.g. 'node_feature' -> feature dimension;
                     'graph_label' -> label dimension
        """
        dim_dict = {}
        for key in self._graph_example.keys:
            tensor = self._graph_example[key]
            if not torch.is_tensor(tensor):
                continue
            if tensor.ndim == 1:
                dim_dict[key] = 1
            elif tensor.ndim == 2:
                dim_dict[key] = tensor.size()[-1]
            else:
                raise ValueError('Dimension of tensor {} exceeds 2.'.format(key))
        return dim_dict

    def _split_transductive(
            self, split_ratio: List[float],
            split_types: List[str] = None) -> Union[List[Graph], List[HeteroGraph]]:
        r"""
        Split the dataset assuming training process is transductive.

        Args:
            split_ratio: number of data splitted into train, validation
                (and test) set.

        Returns:
            list: A list of 3 (2) lists of :class:`deepsnap.graph.Graph` object corresponding
            to train, validation (and test) set.
        """
        if self.task == 'graph':
            raise ValueError('Graph prediction task cannot be transductive')

        # a list of split graphs
        # (e.g. [[train graph, val graph, test graph], ... ])
        graphs_split_list = []
        for graph in self.graphs:
            if type(graph) == Graph:
                graphs_split = graph.split(self.task, split_ratio)
            elif type(graph) == HeteroGraph:
                graphs_split = graph.split(
                                task=self.task,
                                split_types=split_types,
                                split_ratio=split_ratio,
                                edge_split_mode=self.edge_split_mode)
            else:
                raise TypeError('element in self.graphs of unexpected type')
            if self.task == 'link_pred' and \
                    self.edge_train_mode == 'disjoint':
                if type(graphs_split[0]) == Graph:
                    graphs_split[0] = graphs_split[0].split_link_pred(
                        self.edge_message_ratio)[1]
                elif type(graphs_split[0]) == HeteroGraph:
                    graphs_split[0] = graphs_split[0].split_link_pred(
                        split_types=split_types,
                        split_ratio=self.edge_message_ratio,
                        edge_split_mode=self.edge_split_mode)[1]
                else:
                    raise TypeError('element in self.graphs of unexpected type')

            graphs_split_list.append(graphs_split)

        # list of num_splits datasets
        # (e.g. [train dataset, val dataset, test dataset])
        dataset_return = []
        for x in zip(*graphs_split_list):
            dataset_current = copy.copy(self)
            dataset_current.graphs = list(x)
            if self.task == 'link_pred':
                for graph_temp in dataset_current.graphs:
                    if type(graph_temp) == Graph:
                        graph_temp._create_neg_sampling(
                            self.edge_negative_sampling_ratio)
                    elif type(graph_temp) == HeteroGraph:
                        graph_temp._create_neg_sampling(
                            negative_sampling_ratio=self.edge_negative_sampling_ratio,
                            split_types=split_types)

            dataset_return.append(dataset_current)

        # resample negatives for train split (only for link prediction)
        dataset_return[0]._resample_negatives = True

        return dataset_return

    def _split_inductive(
            self,
            split_ratio: List[float],
            split_types: List[str] = None) -> Union[List[Graph], List[HeteroGraph]]:
        r"""
        Split the dataset assuming training process is inductive.

        Args:
            split_ratio: number of data splitted into train, validation
                (and test) set.

        Returns:
            List[Graph]: a list of 3 (2) lists of graph object corresponding to train, validation (and test) set.
        """
        num_graphs = len(self.graphs)
        if (num_graphs < len(split_ratio)):
            raise ValueError('in _split_inductive num of graphs are smaller than the number of splitted parts')

        self._shuffle()
        # a list of num_splits list of graphs
        # (e.g. [train graphs, val graphs, test graphs])
        split_graphs = []
        split_offset = 0

        # perform `secure split` s.t. guarantees all splitted graph list
        # contains at least one graph.
        for i, split_ratio_i in enumerate(split_ratio):
            if i != len(split_ratio) - 1:
                num_split_i = 1 + int(split_ratio_i * (num_graphs - len(split_ratio)))
                split_graphs.append(
                    self.graphs[split_offset: split_offset + num_split_i])
                split_offset += num_split_i
            else:
                split_graphs.append(self.graphs[split_offset:])

        # create objectives for link_pred task
        if self.task == 'link_pred':
            # if disjoint, this will split all graph's edges into 2:
            # message passing and objective edges
            # which is returned by the [1] of the split graphs
            if self.edge_train_mode == 'disjoint':
                split_start = 0
            # in all mode, train graph has all edges used for both
            # message passing and objective.
            elif self.edge_train_mode == 'all':
                split_start = 1
            for i in range(split_start, len(split_graphs)):
                for j in range(len(split_graphs[i])):
                    if type(split_graphs[i][j]) == Graph:
                        split_graphs[i][j] = \
                            split_graphs[i][j].split_link_pred(
                                self.edge_message_ratio)[1]
                    elif type(split_graphs[i][j]) == HeteroGraph:
                        split_graphs[i][j] = \
                            split_graphs[i][j].split_link_pred(
                                split_types, self.edge_message_ratio, self.edge_split_mode)[1]
                    else:
                        raise TypeError('element in self.graphs of unexpected type')

        # list of num_splits datasets
        dataset_return = []
        for graphs in split_graphs:
            dataset_current = copy.copy(self)
            dataset_current.graphs = graphs
            if self.task == 'link_pred':
                for graph_temp in dataset_current.graphs:
                    if type(graph_temp) == Graph:
                        graph_temp._create_neg_sampling(
                            self.edge_negative_sampling_ratio)
                    elif type(graph_temp) == HeteroGraph:
                        graph_temp._create_neg_sampling(
                            negative_sampling_ratio=self.edge_negative_sampling_ratio,
                            split_types=split_types)
            dataset_return.append(dataset_current)

        # resample negatives for train split (only for link prediction)
        dataset_return[0]._resample_negatives = True

        return dataset_return

    def split(
            self,
            transductive: bool = True,
            split_ratio: List[float] = None,
            split_types: Union[str, List[str]] = None) -> Union[List[Graph], List[HeteroGraph]]:
        r""" Split datasets into train, validation (and test) set.

        Args:
            transductive: whether the training process is transductive
                or inductive. Inductive split is always used for graph-level tasks (
                self.task == 'graph').
            split_ratio: number of data splitted into train, validation
                (and test) set.

        Returns:
            list: a list of 3 (2) lists of :class:`deepsnap.graph.Graph` objects corresponding to train, validation (and test) set.
        """
        if self.graphs is None:
            raise RuntimeError('Split is not supported for on-the-fly dataset.'
                               'Construct different on-the-fly datasets for train, val and test.'
                               'Or perform split at batch level.')
        if split_ratio is None:
            split_ratio = [0.8, 0.1, 0.1]
        if not isinstance(split_ratio, list):
            raise TypeError('Split ratio must be a list.')
        if (len(split_ratio) > 3):
            raise ValueError(
                'Split ratio must contain less than or equal to three values.')
        if not math.isclose(sum(split_ratio), 1.0):
            raise ValueError('Split ratio must sum up to 1.')
        if not all(isinstance(split_ratio_i, float) for
                   split_ratio_i in split_ratio):
            raise TypeError('Split ratio must contain all floats')
        if not all(split_ratio_i > 0 for split_ratio_i in split_ratio):
            raise ValueError('Split ratio must contain all positivevalues.')

        # Support for split_types passed in as string instead of list of strings
        if isinstance(split_types, str):
            split_types = [split_types]

        # store the most recent split types
        self._split_types = split_types

        # list of num_splits datasets
        dataset_return = []
        if transductive and self.task != 'graph':
            dataset_return = \
                self._split_transductive(split_ratio, split_types)
        else:
            dataset_return = \
                self._split_inductive(split_ratio, split_types)

        return dataset_return

    def resample_disjoint(self):
        r""" Resample disjoint edge split of message passing and objective links.

        Note that if apply_transform (on the message passing graph)
        was used before this resampling, it needs to be
        re-applied, after resampling, to update some of the edges that were in objectives.
        """
        if self.graphs is None:
            raise RuntimeError('Resampling disjoint is not needed for on-the-fly dataset. '
                               'Split the on-the-fly data as the batch arrives.')
        for graph in self.graphs:
            graph.resample_disjoint(self.edge_message_ratio)

    def _reset_cache(self):
        r"""
        Resets internal cache for graph examples, num_node_labels etc.
        """
        self._num_node_labels = None
        self._num_nodes = None
        self._num_edge_labels = None
        self._num_edges = None
        self._num_graph_labels = None
        if self.graphs is None:
            self._graph_example = self.generator.generate()
            if not isinstance(self._graph_example, Graph):
                self._graph_example = Graph(self._graph_example)
        else:
            self._graph_example = self.graphs[0]

    def apply_transform(self, transform, update_tensor: bool = True,
                        update_graph: bool = False,
                        deep_copy: bool = False, **kwargs):
        r"""
        Applies a transformation to each graph object in parallel by first
        calling `to_data_list`, applying the transform, and then perform
        re-batching again to a `GraphDataset`.

        Args:
            transform: user-defined transformation function.
            update_tensor: whether request the Graph object remain unchanged.
            kwargs: parameters used in transform function in Graph object.
        """
        # currently does not support transform for on-the-fly dataset
        if self.graphs is None:
            raise ValueError('On-the-fly datasets do not support transform.'
                    'Transform can be done at the batch level.')
        # TODO: parallel apply
        new_dataset = copy.copy(self)
        new_dataset.graphs = [
            graph.apply_transform(transform, update_tensor, update_graph,
            deep_copy, **kwargs) for graph in self.graphs]
        # update example graph used for num_node_features etc.
        new_dataset._reset_cache()
        return new_dataset

    def filter(self, filter_fn, deep_copy: bool = False, **kwargs):
        r""" Filter the dataset, discarding graph data G where filter_fn(G) is False.

        GraphDataset.apply_transform is an analog of python map in graph dataset, while
        GraphDataset.filter is an analog of python filter.

        Args:
            filter_fn: user-defined filter function that returns True (keep) or
                False (discard) for graph object in this dataset.
            deep_copy: whether to deep copy all graph objects in the returned list.
            kwargs: parameters used in the filter function.

        Returns:
            A new dataset where graphs are filtered by the given filter function.
        """
        # currently does not support filter for on-the-fly dataset
        if self.graphs is None:
            raise ValueError('On-the-fly datasets do not support transform.'
                    'Filter can be done at the batch level.')
        new_dataset = copy.copy(self)
        new_dataset.graphs = [
            graph for graph in self.graphs if filter_fn(graph, **kwargs)]
        # update example graph used for num_node_features etc.
        new_dataset._reset_cache()
        return new_dataset

    def to(self, device):
        r"""
        Transfer Graph object in the graphs to specified device.

        Args:
            device: Specified device name
        """
        if self.graphs is None:
            self.otf_device = device
        else:
            for graph in self.graphs:
                graph.to(device)

    def _shuffle(self):
        r"""
        shuffle Graph object in graphs.
        """
        if self.graphs is not None:
            random.shuffle(self.graphs)

    @staticmethod
    def pyg_to_graphs(dataset, verbose: bool = False, fixed_split: bool = False) -> List[Graph]:
        r"""
        Transform a torch_geometric.data.Dataset object to a list of Graph object.

        Args:
            dataset: a torch_geometric.data.Dataset object.
            verbose: if print verbose warning
            fixed_split: if load fixed data split from PyG dataset

        Returns:
            list: A list of :class:`deepsnap.graph.Graph` object.
        """
        if fixed_split:
            graphs = [Graph.pyg_to_graph(data, verbose=verbose, fixed_split=True) for data in dataset]
            graphs_split = [[graph] for graph in graphs[0]]
            return graphs_split
        else:
            return [Graph.pyg_to_graph(data, verbose=verbose) for data in dataset]

    @staticmethod
    def list_to_graphs(G_list) -> List[Graph]:
        r"""
        Transform a list of networkx data object to a list of Graph object.

        Args:
            G_list: a list of networkx data object.

        Returns:
            list: A list of :class:`deepsnap.graph.Graph` object.
        """
        return [Graph(G) for G in G_list]

    def __getitem__(self, idx: int) -> Union[Graph, List[Graph]]:
        r"""
        Takes in an integer (or a list of integers)
        returns a single Graph object (a subset of graphs).

        Args:
            idx: index to be selected from graphs.

        Returns:
            Union[:class:`deepsnap.graph.Graph`, List[:class:`deepsnap.graph.Graph`]]: A single
            :class:`deepsnap.graph.Graph` object or subset of :class:`deepsnap.graph.Graph` objects.
        """
        if self.task == 'link_pred' and self._resample_negatives:
            # resample negative examples
            for graph in self.graphs:
                if type(graph) == Graph:
                    graph._create_neg_sampling(
                        self.edge_negative_sampling_ratio, resample=True)
                elif type(graph) == HeteroGraph:
                    graph._create_neg_sampling(
                        self.edge_negative_sampling_ratio, split_types=self._split_types, resample=True)

        # TODO: add the hetero graph equivalent of these functions ?
        if self.graphs is None:
            graph = self.generator.generate()
            if not isinstance(graph, Graph):
                graph = Graph(graph)
            # generated an networkx graph
            if self.otf_device is not None:
                graph.to(self.otf_device)
            return graph
        elif isinstance(idx, int):
            return self.graphs[idx]
        else:
            return self._index_select(idx)

    def _index_select(self, idx: int) -> List[Graph]:
        r"""
        Takes in a list of integers, returns a subset of graphs
        corresponding to the list of integers.

        Args:
            idx: index to be selected from graphs.

        Returns:
            List[Graph]: a single Graph object or subset of graphs.
        """
        if self.graphs is None:
            # _index_select is only called when self.graphs is not None
            raise NotImplementedError('Index select is not available for on-the-fly dataset.')
        if isinstance(idx, slice):
            dataset = copy.copy(self)
            dataset.graphs = self.graphs[idx]
            return dataset
        elif torch.is_tensor(idx):
            if idx.dtype == torch.long:
                return self._index_select(idx.tolist())
            elif idx.dtype == torch.bool or idx.dtype == torch.uint8:
                return self._index_select(idx.nonzero().flatten().tolist())
        elif isinstance(idx, list) or isinstance(idx, tuple):
            dataset = copy.copy(self)
            dataset.graphs = [self.graphs[i] for i in idx]
            return dataset
        else:
            raise IndexError(
                'Only integers, slices (`:`), list, tuples, and long or bool '
                'tensors are valid indices (got {}).'.format(type(idx).__name__))

    def __repr__(self) -> str:  # pragma: no cover
        descriptor = len(self) if self.graphs is not None else self.generator.__class__
        return f'{self.__class__.__name__}({descriptor})'



import torch
import unittest
import numpy as np
from tests.utils import *
from deepsnap.graph import Graph
from torch_geometric.datasets import Planetoid


class TestGraph(unittest.TestCase):

    def test_add_feature_nx(self):
        G, x, y, edge_x, edge_y, edge_index, graph_x, graph_y = \
            simple_networkx_graph()
        Graph.add_edge_attr(G, "edge_feature", edge_x)
        Graph.add_edge_attr(G, "edge_label", edge_y)
        Graph.add_node_attr(G, "node_feature", x)
        Graph.add_node_attr(G, "node_label", y)
        Graph.add_graph_attr(G, "graph_feature", graph_x)
        Graph.add_graph_attr(G, "graph_label", graph_y)

        self.assertEqual(len(G.edges.data()), edge_index.shape[1])
        for item in G.edges.data():
            self.assertEqual('edge_feature' in item[2], True)
            self.assertEqual('edge_label' in item[2], True)
            self.assertEqual(len(item[2]['edge_feature']), 2)
            self.assertEqual(type(item[2]['edge_label'].item()), int)

        for item in G.nodes.data():
            self.assertEqual('node_feature' in item[1], True)
            self.assertEqual('node_label' in item[1], True)
            self.assertEqual(len(item[1]['node_feature']), 2)
            self.assertEqual(type(item[1]['node_label'].item()), int)

        self.assertEqual(G.graph.get("graph_feature").
                         eq(graph_x).sum().item(), 2)
        self.assertEqual(G.graph.get("graph_label").
                         eq(graph_y).sum().item(), 1)

    def test_graph_basics(self):
        G, x, y, edge_x, edge_y, edge_index, graph_x, graph_y = \
            simple_networkx_graph()
        Graph.add_edge_attr(G, "edge_feature", edge_x)
        Graph.add_edge_attr(G, "edge_label", edge_y)
        Graph.add_node_attr(G, "node_feature", x)
        Graph.add_node_attr(G, "node_label", y)
        Graph.add_graph_attr(G, "graph_feature", graph_x)
        Graph.add_graph_attr(G, "graph_label", graph_y)
        dg = Graph(G)
        self.assertTrue(dg.is_directed())
        self.assertEqual(dg.is_undirected(), False)
        self.assertEqual(len(dg), 10)
        for item in ['G', 'node_feature', 'node_label', 'edge_feature',
                     'edge_label', 'graph_feature', 'graph_label',
                     'edge_index', 'edge_label_index', 'node_label_index']:
            self.assertEqual(item in dg, True)
        self.assertEqual(len([key for key in dg]), 10)

    def test_graph_property_edge_case(self):
        G_1, x, y, edge_x, edge_y, edge_index, graph_x, graph_y = \
            simple_networkx_graph()
        dg_1 = Graph(G_1)
        self.assertEqual(dg_1.num_nodes, G_1.number_of_nodes())
        self.assertEqual(dg_1.num_edges, G_1.number_of_edges())
        self.assertEqual(dg_1.num_node_features, 0)
        self.assertEqual(dg_1.num_edge_features, 0)
        self.assertEqual(dg_1.num_graph_features, 0)
        self.assertEqual(dg_1.num_node_labels, 0)
        self.assertEqual(dg_1.num_edge_labels, 0)
        self.assertEqual(dg_1.num_graph_labels, 0)

        G_2, x, y, edge_x, edge_y, edge_index, graph_x, graph_y = \
            simple_networkx_graph()
        Graph.add_edge_attr(G_2, "edge_label", edge_y.type(torch.FloatTensor))
        Graph.add_node_attr(G_2, "node_label", y.type(torch.FloatTensor))
        Graph.add_graph_attr(G_2, "graph_label",
                             graph_y.type(torch.FloatTensor))

        dg_2 = Graph(G_2)
        self.assertEqual(dg_2.num_node_labels, 1)
        self.assertEqual(dg_2.num_edge_labels, 1)
        self.assertEqual(dg_2.num_graph_labels, 1)

    def test_graph_property_general(self):
        G, x, y, edge_x, edge_y, edge_index, graph_x, graph_y = \
            simple_networkx_graph()
        Graph.add_edge_attr(G, "edge_feature", edge_x)
        Graph.add_edge_attr(G, "edge_label", edge_y)
        Graph.add_node_attr(G, "node_feature", x)
        Graph.add_node_attr(G, "node_label", y)
        Graph.add_graph_attr(G, "graph_feature", graph_x)
        Graph.add_graph_attr(G, "graph_label", graph_y)

        dg = Graph(G)
        self.assertEqual(dg.keys,
                         ['G', 'node_feature', 'node_label', 'edge_feature',
                          'edge_label', 'graph_feature', 'graph_label',
                          'edge_index', 'edge_label_index',
                          'node_label_index'])
        self.assertEqual(dg.num_nodes, G.number_of_nodes())
        self.assertEqual(dg.num_edges, G.number_of_edges())
        self.assertEqual(dg.num_node_features, 2)
        self.assertEqual(dg.num_edge_features, 2)
        self.assertEqual(dg.num_graph_features, 2)
        self.assertEqual(dg.num_node_labels, np.max(y.data.numpy()) + 1)
        self.assertEqual(dg.num_edge_labels, np.max(edge_y.data.numpy()) + 1)
        self.assertEqual(dg.num_graph_labels, np.max(graph_y.data.numpy()) + 1)

    def test_pyg_to_graph(self):
        pyg_dataset = Planetoid('./cora', 'Cora')
        dg = Graph.pyg_to_graph(pyg_dataset[0])
        pyg_data = pyg_dataset[0]
        self.assertEqual(pyg_data.num_nodes, dg.num_nodes)
        self.assertEqual(pyg_data.is_directed(), dg.is_directed())
        self.assertEqual(pyg_data.num_edges / 2, dg.num_edges)
        self.assertTrue(dg.num_node_features == pyg_data.x.shape[1])
        self.assertTrue(dg.num_node_labels == torch.max(pyg_data.y).item() + 1)
        self.assertTrue(dg.edge_index.shape == pyg_data.edge_index.shape)
        keys = ['G', 'node_feature', 'node_label', 'edge_index',
                'edge_label_index', 'node_label_index']
        self.assertTrue(tuple(dg.keys) == tuple(keys))

    def test_clone(self):
        G, x, y, edge_x, edge_y, edge_index, graph_x, graph_y = \
            simple_networkx_graph()
        Graph.add_edge_attr(G, "edge_feature", edge_x)
        Graph.add_edge_attr(G, "edge_label", edge_y)
        Graph.add_node_attr(G, "node_feature", x)
        Graph.add_node_attr(G, "node_label", y)
        Graph.add_graph_attr(G, "graph_feature", graph_x)
        Graph.add_graph_attr(G, "graph_label", graph_y)

        dg = Graph(G)
        dg1 = dg.clone()
        self.assertEqual(dg.num_nodes, dg1.num_nodes)
        self.assertEqual(dg.num_edges, dg1.num_edges)
        self.assertEqual(dg.num_node_features, dg1.num_node_features)
        self.assertEqual(dg.num_edge_features, dg1.num_edge_features)
        self.assertEqual(dg.num_node_labels, dg1.num_node_labels)
        self.assertEqual(dg.num_edge_labels, dg1.num_edge_labels)
        self.assertTrue(not id(dg.G) == id(dg1.G))
        self.assertTrue(tuple(dg.keys) == tuple(dg1.keys))

    def test_split_edge_case(self):
        G, x, y, edge_x, edge_y, edge_index, graph_x, graph_y = \
            simple_networkx_graph()
        dg = Graph(G)

        dg_node = dg.split()
        dg_num_nodes_reduced = dg.num_nodes - 3
        self.assertEqual(
            dg_node[0].node_label_index.shape[0], 1 + int(dg_num_nodes_reduced * 0.8))
        self.assertEqual(
            dg_node[1].node_label_index.shape[0], 1 + int(dg_num_nodes_reduced * 0.1))
        self.assertEqual(
            dg_node[2].node_label_index.shape[0], dg.num_nodes - 2 -
            int(dg_num_nodes_reduced * 0.8) - int(dg_num_nodes_reduced * 0.1))

        dg_edge = dg.split(task='edge')
        dg_num_edges_reduced = dg.num_edges - 3
        edge_0 = 1 + int(dg_num_edges_reduced * 0.8)
        edge_1 = 1 + int(dg_num_edges_reduced * 0.1)
        edge_2 = dg.num_edges - edge_0 - edge_1
        self.assertEqual(dg_edge[0].edge_label_index.shape[1], edge_0)
        self.assertEqual(dg_edge[1].edge_label_index.shape[1], edge_1)
        self.assertEqual(dg_edge[2].edge_label_index.shape[1], edge_2)

        dg_link = dg.split(task='link_pred')
        dg_num_edges_reduced = dg.num_edges - 3
        edge_0 = 1 + int(dg_num_edges_reduced * 0.8)
        edge_1 = 1 + int(dg_num_edges_reduced * 0.1)
        edge_2 = dg.num_edges - edge_0 - edge_1
        self.assertEqual(dg_link[0].edge_label_index.shape[1], edge_0)
        self.assertEqual(dg_link[1].edge_label_index.shape[1], edge_1)
        self.assertEqual(dg_link[2].edge_label_index.shape[1], edge_2)

    def test_split(self):
        pyg_dataset = Planetoid('./cora', 'Cora')
        dg = Graph.pyg_to_graph(pyg_dataset[0])

        dg_node = dg.split()
        dg_num_nodes_reduced = dg.num_nodes - 3
        self.assertEqual(
            dg_node[0].node_label_index.shape[0], 1 + int(dg_num_nodes_reduced * 0.8))
        self.assertEqual(
            dg_node[1].node_label_index.shape[0], 1 + int(dg_num_nodes_reduced * 0.1))
        self.assertEqual(
            dg_node[2].node_label_index.shape[0], dg.num_nodes - 2 -
            int(dg_num_nodes_reduced * 0.8) - int(dg_num_nodes_reduced * 0.1))

        dg_edge = dg.split(task='edge')
        dg_num_edges_reduced = dg.num_edges - 3
        edge_0 = 2 * (1 + int(dg_num_edges_reduced * 0.8))
        edge_1 = 2 * (1 + int(dg_num_edges_reduced * 0.1))
        edge_2 = dg.num_edges * 2 - edge_0 - edge_1
        self.assertEqual(dg_edge[0].edge_label_index.shape[1], edge_0)
        self.assertEqual(dg_edge[1].edge_label_index.shape[1], edge_1)
        self.assertEqual(dg_edge[2].edge_label_index.shape[1], edge_2)

        dg_link = dg.split(task='link_pred')
        dg_num_edges_reduced = dg.num_edges - 3
        edge_0 = 2 * (1 + int(dg_num_edges_reduced * 0.8))
        edge_1 = 2 * (1 + int(dg_num_edges_reduced * 0.1))
        edge_2 = dg.num_edges * 2 - edge_0 - edge_1
        self.assertEqual(dg_link[0].edge_label_index.shape[1], edge_0)
        self.assertEqual(dg_link[1].edge_label_index.shape[1], edge_1)
        self.assertEqual(dg_link[2].edge_label_index.shape[1], edge_2)

        for message_ratio in [0.1, 0.2, 0.4, 0.8]:
            dg_link_resample = dg_link[0].clone().\
                resample_disjoint(message_ratio=message_ratio)
            positive_edge_num = \
                int(0.5 * dg_link[0].clone().edge_label_index.shape[1])
            self.assertEqual(dg_link_resample.edge_label_index.shape[1],
                             2 * (positive_edge_num - 1 - int(message_ratio * (positive_edge_num - 2))))

        for split_ratio in [[0.1, 0.4, 0.5], [0.4, 0.3, 0.3], [0.7, 0.2, 0.1]]:
            dg_link_custom = \
                dg.split(task='link_pred', split_ratio=split_ratio)
            dg_num_edges_reduced = dg.num_edges - 3
            edge_0 = 2 * (1 + int(dg_num_edges_reduced * split_ratio[0]))
            self.assertEqual(dg_link_custom[0].
                             edge_label_index.shape[1], edge_0)
            edge_1 = \
                (1 + int(split_ratio[0] * dg_num_edges_reduced) + 1 + int(split_ratio[1] * dg_num_edges_reduced)) * 2 - edge_0
            self.assertEqual(dg_link_custom[1].edge_label_index.shape[1],
                             edge_1)
            edge_2 = dg.num_edges * 2 - edge_0 - edge_1
            self.assertEqual(dg_link_custom[2].edge_label_index.shape[1],
                             edge_2)

    def test_transform(self):
        G, x, y, edge_x, edge_y, edge_index, graph_x, graph_y = \
            simple_networkx_graph()
        Graph.add_edge_attr(G, "edge_feature", edge_x)
        Graph.add_edge_attr(G, "edge_label", edge_y)
        Graph.add_node_attr(G, "node_feature", x)
        Graph.add_node_attr(G, "node_label", y)
        Graph.add_graph_attr(G, "graph_feature", graph_x)
        Graph.add_graph_attr(G, "graph_label", graph_y)
        dg = Graph(G)

        dg_edge_feature = dg.edge_feature.clone()
        dg_node_feature = dg.node_feature.clone()
        dg_graph_feature = dg.graph_feature.clone()

        dg.apply_tensor(
            lambda x: x, 'edge_feature', 'node_feature', 'graph_feature')
        self.assertTrue(torch.all(dg_edge_feature.eq(dg.edge_feature)))
        self.assertTrue(torch.all(dg_node_feature.eq(dg.node_feature)))
        self.assertTrue(torch.all(dg_graph_feature.eq(dg.graph_feature)))

        dg.apply_tensor(
            lambda x: x + 10, 'edge_feature', 'node_feature', 'graph_feature')
        self.assertFalse(torch.all(dg_edge_feature.eq(dg.edge_feature)))
        self.assertFalse(torch.all(dg_node_feature.eq(dg.node_feature)))
        self.assertFalse(torch.all(dg_graph_feature.eq(dg.graph_feature)))

        dg.apply_tensor(
            lambda x: x + 100, 'edge_feature', 'node_feature', 'graph_feature')
        self.assertTrue(
            torch.all(dg.edge_feature.eq(dg_edge_feature + 10 + 100)))
        self.assertTrue(
            torch.all(dg.node_feature.eq(dg_node_feature + 10 + 100)))
        self.assertTrue(
            torch.all(dg.graph_feature.eq(dg_graph_feature + 10 + 100)))

        dg.apply_tensor(
            lambda x: x * 2, 'edge_feature', 'node_feature', 'graph_feature')
        self.assertTrue(
            torch.all(dg.edge_feature.eq((dg_edge_feature + 10 + 100) * 2)))
        self.assertTrue(
            torch.all(dg.node_feature.eq((dg_node_feature + 10 + 100) * 2)))
        self.assertTrue(
            torch.all(dg.graph_feature.eq((dg_graph_feature + 10 + 100) * 2)))

    def test_repr(self):
        G, x, y, edge_x, edge_y, edge_index, graph_x, graph_y = \
            simple_networkx_graph()
        Graph.add_edge_attr(G, "edge_feature", edge_x)
        Graph.add_edge_attr(G, "edge_label", edge_y)
        Graph.add_node_attr(G, "node_feature", x)
        Graph.add_node_attr(G, "node_label", y)
        Graph.add_graph_attr(G, "graph_feature", graph_x)
        Graph.add_graph_attr(G, "graph_label", graph_y)
        dg = Graph(G)
        self.assertEqual(
            repr(dg),
            'Graph(G=[], edge_feature=[17, 2], '
            'edge_index=[2, 17], edge_label=[17], edge_label_index=[2, 17], '
            'graph_feature=[1, 2], graph_label=[1], node_feature=[10, 2], '
            'node_label=[10], node_label_index=[10])')


if __name__ == '__main__':
    unittest.main()



import math
import torch
import unittest
import networkx as nx
from deepsnap.graph import Graph
from deepsnap.batch import Batch
from deepsnap.dataset import GraphDataset
from torch.utils.data import DataLoader
from torch_geometric.datasets import TUDataset
from copy import deepcopy
from tests.utils import *


class TestBatch(unittest.TestCase):

    def test_batch_basic(self):
        G, x, y, edge_x, edge_y, edge_index, graph_x, graph_y = \
            simple_networkx_graph()
        Graph.add_edge_attr(G, "edge_feature", edge_x)
        Graph.add_edge_attr(G, "edge_label", edge_y)
        Graph.add_node_attr(G, "node_feature", x)
        Graph.add_node_attr(G, "node_label", y)
        Graph.add_graph_attr(G, "graph_feature", graph_x)
        Graph.add_graph_attr(G, "graph_label", graph_y)
        H = deepcopy(G)
        graphs = GraphDataset.list_to_graphs([G, H])
        batch = Batch.from_data_list(graphs)
        self.assertEqual(batch.num_graphs, 2)
        self.assertEqual(
            len(batch.node_feature), 2 * len(graphs[0].node_feature))

    def test_torch_dataloader_collate(self):
        # graph classification example
        pyg_dataset = TUDataset('./enzymes', 'ENZYMES')
        graphs = GraphDataset.pyg_to_graphs(pyg_dataset)
        dataset = GraphDataset(graphs, task="graph")
        train_batch_num = math.ceil(len(dataset) * 0.8 / 32)
        test_batch_num = math.ceil(len(dataset) * 0.1 / 32)
        val_batch_num = math.ceil(len(dataset) * 0.1 / 32)
        datasets = {}
        datasets['train'], datasets['val'], datasets['test'] = \
            dataset.split(transductive=False, split_ratio=[0.8, 0.1, 0.1])
        dataloaders = {split: DataLoader(
                    dataset, collate_fn=Batch.collate(),
                    batch_size=32, shuffle=True)
                    for split, dataset in datasets.items()}
        self.assertEqual(len(dataloaders['train']), train_batch_num)
        self.assertEqual(len(dataloaders['val']), test_batch_num)
        self.assertEqual(len(dataloaders['test']), val_batch_num)
        for i, data in enumerate(dataloaders['train']):
            if i != len(dataloaders['train']) - 1:
                self.assertEqual(data.num_graphs, 32)
        for i, data in enumerate(dataloaders['val']):
            if i != len(dataloaders['val']) - 1:
                self.assertEqual(data.num_graphs, 32)
        for i, data in enumerate(dataloaders['test']):
            if i != len(dataloaders['test']) - 1:
                self.assertEqual(data.num_graphs, 32)

    def test_collate_batch_nested(self):
        dims = [2, 3]
        G_sizes = [10, 5]
        G_list = []
        for i, size in enumerate(G_sizes):
            G = Graph()
            G.G = nx.complete_graph(i+1)
            G.node_property = {'node_prop0': torch.ones(size, dims[0]) * i,
                               'node_prop1': torch.ones(size, dims[1]) * i}
            G_list.append(G)
        batch = Batch.from_data_list(G_list)

        self.assertEqual(batch.num_graphs, 2)
        self.assertEqual(
            batch.node_property['node_prop0'].size(0), sum(G_sizes))

    def test_unbatch_nested(self):
        dims = [2, 3]
        G_sizes = [10, 5]
        G_list = []
        for i, size in enumerate(G_sizes):
            G = Graph()
            G.G = nx.complete_graph(i+1)
            G.node_property = {'node_prop0': torch.ones(size, dims[0]) * i,
                               'node_prop1': torch.ones(size, dims[1]) * i}
            G_list.append(G)
        batch = Batch.from_data_list(G_list)

        # reconstruct graph list
        G_list_recon = batch.to_data_list()
        self.assertEqual(
            G_list_recon[0].node_property['node_prop0'].size(0), 10)
        self.assertEqual(
            G_list_recon[0].node_property['node_prop0'].size(1), 2)
        self.assertEqual(
            G_list_recon[1].node_property['node_prop1'].size(0), 5)
        self.assertEqual(
            G_list_recon[1].node_property['node_prop1'].size(1), 3)


if __name__ == '__main__':
    unittest.main()
